//package directSpread;
//import java.awt.Shape;
//import java.util.Arrays;
//import java.util.Random;
//
//import directSpread.utility.Operator;
//import directSpread.utility.Solution;
//import directSpread.utility.Variable;
//
///**
// * Mutates a coordinate by moving it away from its nearest neighbours.
// *  The vectors between the coordinate and its nearest neighbours are combined
// *  to create a displacement vector; the direction of this vector is modified by 
// *  a normal distribution and the distance set with a normal distribution based 
// *  on a default length.
// *  
// *  Operates on a single coordinate per entry.
// */
//public class CoordinateUniformCrossover implements Operator{
//
//	private Random random = new Random();
//	private int nearestNeighbours;
//	private double directionSigma;
//	private double distanceMu;
//	private double distanceSigma;
//	private Shape bounds;
//	private double minimumDistance;
//	private double probability;
//	private double reversalProbability;
//	
//	private int reduction = 8;  //Magic number used to force displaced coordinate into a legal value. 
//
//	
//	
//	/**
//	 * Creates an instance of the Vector-Based Movement Mutation for coordinate systems.
//	 * 
//	 * @param probability  The probability that this mutation will occur.
//	 * @param nearestNeighbours  The number of nearest neighbours to use to construct a
//	 *  displacement vector. This must be greater than 0.
//	 * @param bounds  The area the nodes may occupy.
//	 * @param minimumDistance  The minimumDistance between two nodes.
//	 * @param directionSigma  The standard deviation for the direction.
//	 * @param distanceMu  The mean distance to move.
//	 * @param distanceSigma  The standard deviation for the distance.
//	 * @param reversalProbability  The probability the displacement will be reversed.
//	 */
//	public CoordinateUniformCrossover(double probability, int nearestNeighbours, Shape bounds, double minimumDistance,
//			double directionSigma, double distanceMu, double distanceSigma){
//		if (nearestNeighbours < 0){
//			System.err.println("Error in VectorBasedMovement constructor: nearestNeighbours cannot be < 0.");
//			System.exit(1);
//		}
//		this.probability = probability;
//		this.nearestNeighbours = nearestNeighbours;
//		this.bounds = bounds;
//		this.minimumDistance = minimumDistance;
//		this.directionSigma = directionSigma;
//		this.distanceMu = distanceMu;
//		this.distanceSigma = distanceSigma;
//		this.reversalProbability = reversalProbability;
//	}
//	
//	/**
//	 * Works on a series of x,y coordinates in the form [x0, y0, x1, y1, ...].
//	 * Selects and mutates a coordinate by moving it away from its nearest neighbours.  
//	 * 
//	 * @param object The solution to mutate. 
//	 * @return  The mutated solution.
//	 */
//	public Object execute(Object object){
//	
//		Solution[] solutions = (Solution[]) object;		
//		doCrossover(probability, solutions[0], solutions[1]);
//		
//		return solutions;
//	}
//	
//	
//	private void doCrossover(double probability, Solution solution1, Solution solution2){
//		
//		Variable[] variables1 = solution1.getDecisionVariables();
//		Variable[] variables2 = solution2.getDecisionVariables();
//		int numNodes = solution1.getDecisionVariables().length/2;
//		
//		for (int currentNode = 0; currentNode < numNodes; currentNode++){
//			double roll = random.nextDouble();
//			Variable x1 = variables1[2*currentNode];
//			Variable y1 = variables1[2*currentNode+1];
//			Variable x2 = variables2[2*currentNode];
//			Variable y2 = variables2[2*currentNode+1];
//			
//			if (roll < probability){
//				double temp = x1.getValue();
//				x1.setValue(x2.getValue());
//				x2.setValue(temp);
//				temp = y1.getValue();
//				y1.setValue(y2.getValue());
//				y2.setValue(temp);
//			}
//		}
//
//		for (int currentNode = 0; currentNode < numNodes; currentNode++){
//			double x1 = variables1[2*currentNode].getValue();
//			double y1 = variables1[2*currentNode+1].getValue();
//			double x2 = variables2[2*currentNode].getValue();
//			double y2 = variables2[2*currentNode+1].getValue();
//			
//			Neighbour[] neighbours1 = computeNeighbours(variables1, currentNode);
//		    while (!isInBounds(variables1[2*currentNode].getValue(), variables1[2*currentNode+1].getValue()) || isTooCloseToOtherNodes(variables1, currentNode, neighbours1)){
//		    	double distance = getDisplacementDistance();
//		    	double angle = getDisplacementDirection(0, 0);
//			    double displacementX = distance * Math.cos(angle);
//			    double displacementY = distance * Math.sin(angle);  
//		    	
//			    variables1[2*currentNode].setValue(x1+displacementX);
//			    variables1[2*currentNode+1].setValue(y1+displacementY);
//				neighbours1 = computeNeighbours(variables1, currentNode);
//		    }
//		    
//			Neighbour[] neighbours2 = computeNeighbours(variables2, currentNode);
//		    while (!isInBounds(variables2[2*currentNode].getValue(), variables2[2*currentNode+1].getValue()) || isTooCloseToOtherNodes(variables2, currentNode, neighbours2)){
//		    	double distance = getDisplacementDistance();
//		    	double angle = getDisplacementDirection(0, 0);
//			    double displacementX = distance * Math.cos(angle);
//			    double displacementY = distance * Math.sin(angle);  
//		    	
//			    variables2[2*currentNode].setValue(x2+displacementX);
//			    variables2[2*currentNode+1].setValue(y2+displacementY);
//				neighbours2 = computeNeighbours(variables2, currentNode);
//		    }
//		    
//		}
//	}
//	
//	/**
//	 * Returns whether the coordinates are within the bounding shape or not.
//	 * @param x  The x coordinate.
//	 * @param y  The y coordinate.
//	 * @return  True if the x,y coordinates are within the bounding shape, false otherwise. 
//	 */
//	private boolean isInBounds(double x, double y){
//		return bounds.contains(x, y);
//	}
//	
//	/**
//	 * Returns whether the specified node is too close to the supplied neighbours.
//	 * @param variables 
//	 * 
//	 * @param variables  The list of all node coordinates. 
//	 * @param currentNode  The specified node.
//	 * @param neighbours  All other nodes.
//	 * @return  True if the specified node is too close to the supplied neighbours, false otherwise. 
//	 */
//	private boolean isTooCloseToOtherNodes(Variable[] variables, int currentNode, Neighbour[] neighbours){
//		if(neighbours.length == 0) return false;
//		
//    	if (determineDistance(variables, currentNode, neighbours[0].getNeighbourIndex()) < minimumDistance){
//
////       	if (neighbours[0].getDistanceFromNode() < minimumDistance){
//    		//Reset variables.
//    	   return true;
//    	}else{
//    		return false;
//    	}
//	}
//	
//	/**
//	 * Returns a direction based on the supplied vector. Uses a normal
//	 *  distribution if the length of the supplied vector is more than 0,
//	 *  otherwise returns a completely random direction.
//	 * 
//	 * @param displacementX  The x displacement of the vector.
//	 * @param displacementY  The y displacement of the vector.
//	 * @return  The new direction in radians.
//	 */
//	private double getDisplacementDirection(double displacementX, double displacementY){
//		
//		double distance = Math.sqrt(displacementX*displacementX + displacementY*displacementY);
//		//If distance is zero, return random direction.
//		if (distance == 0){
//			return random.nextDouble()*Math.PI*2;
//		}
//		double sigma = directionSigma;
//		
//		//Determine mu based of the normalised X and Y values.
//		//Note that acos, asin and atan return values within one pi range.
//		double mu = 0;
//		if (displacementX < 0 && displacementY < 0){
//			mu = Math.atan(displacementY/displacementX)+Math.PI;
//		}else if (displacementY < 0){
//			mu = Math.asin(displacementY/distance);
//		}else{
//			mu = Math.acos(displacementX/distance);
//		}
//		
//		//Obtain the angle based on a normal distribution.
//		double angle = normalDistribution(mu, sigma) % (2*Math.PI);
//		return angle;
//	}
//	
//	/**
//	 * Returns a normally distributed distance based on the global distance
//	 *  mean and standard deviation.
//	 * 
//	 * @return  a normally distributed distance based on the global distance
//	 *  mean and standard deviation.
//	 */
//	private double getDisplacementDistance(){
//		return normalDistribution(distanceMu, distanceSigma);
//	}
//	
//	
//	/**
//	 * Returns a value based on a normal distribution with the supplied mean mu
//	 *  and standard deviation sigma.
//	 * 
//	 * @param mu  The mean for the normal distribution.
//	 * @param sigma  The standard deviation for the normal distribution.
//	 * @return  a value based on a normal distribution with the supplied mean mu
//	 *  and standard deviation sigma.
//	 */
//	private double normalDistribution(double mu, double sigma){
//		return random.nextGaussian()*sigma+mu;
//	}
//	
//	
//	/**
//	 * Assumes a [x0, y0, x1, y1, ...] x,y coordinate set for nodes 0, 1, ...
//	 * 
//	 * @param node1  First node with x at index 2*node1, y at index 2*node1+1
//	 * @param node2  Second node with x at index 2*node2, y at index 2*node2+1
//	 * @return The distance between these two nodes. This is set to
//	 *  Double.MAX_VALUE if the two nodes are the same.
//	 */
//	private double determineDistance(Variable[] variables, int node1, int node2){
//		
//		//
//		if (node1 == node2){
//			return Double.MAX_VALUE;
//		}
//		
//		double xDiff = variables[2*node2].getValue() - variables[2*node1].getValue();
//		double yDiff = variables[2*node2+1].getValue() - variables[2*node1+1].getValue();
//		
//		return Math.sqrt(xDiff*xDiff + yDiff*yDiff);
//	}
//	
//	/**
//	 * Creates an ordered array of the nearest neighbours to a single node.
//	 * The provided node is not included in the array. 
//	 * 
//	 * @param variables  The coordinates for all nodes
//	 * @param node  The node being used to calculate distances to all other nodes.
//	 * @return  An array of indices and distances for neighbours to the closest node.
//	 */
//	private Neighbour[] computeNeighbours(Variable[] variables, int node){
//		
//		int numNodes = variables.length/2;
//		Neighbour neighbours[] = new Neighbour[numNodes-1];
//		
//		//Get distance from this node to all others.
//		for (int neighbour = 0; neighbour < node; neighbour++){
//			neighbours[neighbour] = new Neighbour(neighbour, determineDistance(variables, node, neighbour));
//		}
//		for (int neighbour = node+1; neighbour < numNodes; neighbour++){
//			neighbours[neighbour-1] = new Neighbour(neighbour, determineDistance(variables, node, neighbour));
//		}
//		
//		//Return the indices of the nodes ordered by closest.
//		Arrays.sort(neighbours);
//		return neighbours;
//	}
//	
//	/**
//	 * InnerStorage class allowing easy permutation of neighbour indices and distances.
//	 */
//	class Neighbour implements Comparable<Neighbour>{
//		private int neighbourIndex;
//		private double distanceFromNode;
//		
//		/**
//		 * Storage class constructor; takes in both which index of the node
//		 *  list this neighbour is and how far away it is from the current node. 
//		 * 
//		 * @param neighbourIndex  The index of the node list this neighbour is.
//		 * @param distanceFromNode  The distance from this neighbour to the
//		 *  current node.
//		 */
//		public Neighbour(int neighbourIndex, double distanceFromNode){
//			this.neighbourIndex = neighbourIndex;
//			this.distanceFromNode = distanceFromNode;
//		}
//		
//		/**
//		 * Compares this Neighbour to the supplied by comparing the
//		 *  distances.
//		 *  
//		 *  @param o  The neighbour to compare against this neighbour. 
//		 */
//		public int compareTo(Neighbour o) {
//			return Double.compare(distanceFromNode, o.distanceFromNode);
//		}
//
//		/**
//		 * Returns the index of the node list this neighbour is. 
//		 * @return  The index of the node list this neighbour is.
//		 */
//		public int getNeighbourIndex() {
//			return neighbourIndex;
//		}
//
//		/**
//		 * Returns the distance from this neighbour to the current node.
//		 * @return  The distance from this neighbour to the current node.
//		 */
//		public double getDistanceFromNode() {
//			return distanceFromNode;
//		}
//	}
//}