package directSpread;

import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import cma.CMAEvolutionStrategy;
import cma.CMASolution;
import directSpread.utility.Operator;
import directSpread.utility.Problem;
import directSpread.utility.Solution;
import directSpread.utility.SolutionSet;
import directSpread.utility.WindTurbineProblem;
import windCode.*;


public class TestOpenWind {

    public static void main(String[] args) {
        System.out.println("Test");
        //Algorithm variables
        int maxEvaluations = 10000;
        int populationSize = 1;
        int numberRepetitions = 1;

        //VectorBasedMovementMutation variables
        double probability = 1;
        double reversalProbability = 0.2;
        int nearestNeighbours = 3;
        double directionSigma = Math.PI / 6; // TODO What should this be?
        double distanceMu;// = 5; // TODO What should this be?
        double distanceSigma;// = distanceMu/3; // TODO What should this be?

        //Adaptation variables
        double distanceMuIncrease;
        double distanceSigmaIncrease;
        double distanceMuDecrease;
        double distanceSigmaDecrease;
        double distanceSigmaMax;
        double distanceSigmaMin;
        double distanceMuMax;
        double distanceMuMin;

        //Problem variables
//        int nodes = 32;
        int nodes = 300;


//		double minimumDistance = 3; //Proximity
        Shape bounds = null; //Bounding area.
        double x1 = 0;
        double y1 = 0;
//		double x2 = minimumDistance*(((int)Math.sqrt(nodes)));
//		double y2 = minimumDistance*(((int)Math.sqrt(nodes))-1);
//		double x2 = minimumDistance*30;
//		double y2 = minimumDistance*30;
//		double x2 = 3000;
//		double y2 = 3000;
//        double x2 = 3000;
//        double y2 = 3000;
//		double x2 = 8000; // for 200
//		double y2 = 5000; // for 200
		double x2 = 20000; // for 1000
		double y2 = 10000; // for 1000
//		double x2 = 14000; // for 500
//		double y2 = 7000; // for 500

        int wakeModelScenario = 2;
        String realism = "w"; // scenario number, realism: "[o]ptimal (wakelossfree)"/"consider [w]akeloss"
//        boolean printSeveralOutputsPerDSRun = true;

        boolean displayGUI = true;
        boolean printSeveralOutputsPerDSRepetition = true;
        boolean printSeveralOutputsPerDSRun = true;
        boolean printEveryEvaluation = !true;
        
        
        
        
        //Creating infeasible zones.
        int infeasibleZonesScenario = 0;
        
        int numInfeasibleZones = 3;
        Rectangle2D[] infeasibleZones = new Rectangle2D[numInfeasibleZones];
        infeasibleZones[0] = new Rectangle2D.Double(0, 0, 500, 500);
        infeasibleZones[1] = new Rectangle2D.Double(800, 1700, 300, 1000);
        infeasibleZones[2] = new Rectangle2D.Double(800, 1800, 800, 300);
        
        

        // consider command line arguments, if there are any
        if (args.length == 8) {
            nodes = Integer.parseInt(args[0]);
            maxEvaluations = Integer.parseInt(args[1]);
            wakeModelScenario = Integer.parseInt(args[2]);
            x2 = Integer.parseInt(args[3]);
            y2 = Integer.parseInt(args[4]);
            nearestNeighbours = Integer.parseInt(args[5]);
            numberRepetitions = Integer.parseInt(args[6]);
            displayGUI = Boolean.parseBoolean(args[7]);
            
//            infeasibleZonesScenario = Integer.parseInt(args[8]);
        }

        
        switch (infeasibleZonesScenario) {
            case 0: 
                // use trivial zone as an 'empty scenario' 
                infeasibleZones = new Rectangle2D[]{new Rectangle2D.Double(1, 1, 0, 0)}; 
                break;
            case 1:
                // use default scenario from above
                break;
            default:
                // use trivial zone as an 'empty scenario' 
                infeasibleZones = new Rectangle2D[]{new Rectangle2D.Double(1, 1, 0, 0)}; 
                break;
        }

        Area area = null;
        boolean rectangle = !false;
        if (rectangle) {
            bounds = new Rectangle2D.Double(x1, y1, x2 - x1, y2 - y1);
            area = new RectangularArea(nodes, x2, y2, infeasibleZones);
        } else {
            bounds = new Ellipse2D.Double(x1, y1, x2 - x1, y2 - y1);
        }


        double minimumDistance = 8 * WakeModel.R;


        //Default movement variables
        distanceMu = area.computeInitialStandardDeviation() * 32;
        distanceSigma = distanceMu / 3;
        //End movement variables
        System.out.println("distanceMu=" + distanceMu);

        //Adaptation variables
        distanceMuIncrease = 1.5;
        distanceSigmaIncrease = 1.5;
        distanceMuDecrease = 1.5;
        distanceSigmaDecrease = 1.5;
        distanceSigmaMax = distanceSigma * 2;
        distanceSigmaMin = 1;//distanceSigma/2;
        distanceMuMax = distanceMu * 2;
        distanceMuMin = distanceMu / 2;
        //End adaptation variables


        System.out.println("Algorithm begins");




        WakeModel fitfun = new WakeModel(wakeModelScenario, realism);



        area.printSeveralOutputs = printSeveralOutputsPerDSRun;


        System.out.println("Wind Farm Layout Optimization - DirectSpread");
        
        System.out.println(
        "nodes:"+nodes+", wakeModelScenario:"+wakeModelScenario+
        ", nearestNeighbours:"+nearestNeighbours+", displayGUI:"+displayGUI);
        
        
//        System.out.println("("+commandLineMu+","+commandLineLambda+")-DS, generations:"+commandLineGenerations+", repetitions:"+commandLineRepetitions);
        System.out.println("maxEvaluations:" + maxEvaluations + ", repetitions:" + numberRepetitions);
//        System.out.println("workerThreads="+commandLineWorkerThreads);
//        System.out.println("commandLineComment: '"+commandLineComment+"'");
        System.out.println("infeasible regions:" + infeasibleZones.toString()+ "[zones: "+infeasibleZones.length + ", scenario:" +infeasibleZonesScenario+"]");
        System.out.println();

      System.out.println(area.toString());

      WindTurbineProblem problem = new WindTurbineProblem(nodes,
      		minimumDistance, bounds, area, wakeModelScenario, realism,
      		distanceMu, distanceSigma, directionSigma, 
      		distanceMuIncrease, distanceSigmaIncrease, distanceMuDecrease, distanceSigmaDecrease,
      		distanceSigmaMax, distanceSigmaMin, distanceMuMax, distanceMuMin);
//      WindTurbineProblem problem = new WindTurbineProblem(nodes, minimumDistance, bounds, area, wakeModelScenario, realism);
      
      
      long startTime, stopTime;
      startTime = System.currentTimeMillis();
      Solution bestSolution = runDSSeveralTimes(problem, area, bounds, x2, y2, fitfun, populationSize, populationSize, maxEvaluations,
    		  printEveryEvaluation, printSeveralOutputsPerDSRun, printSeveralOutputsPerDSRepetition, displayGUI, 
          numberRepetitions, directionSigma, distanceMu, distanceSigma, 
          nearestNeighbours, probability, reversalProbability, infeasibleZones
//          ,commandLineWorkerThreads
          );
      stopTime = System.currentTimeMillis();
      double timeNeeded = (stopTime-startTime)/1000d;

      double wakeloss = 1d/fitfun.computeWakelossFreeFitness(area.startingSolution()) - 1d/bestSolution.getObjectives()[0];
      System.out.println("turbines="+area.numberOfTurbines+", energy="+1d/bestSolution.getObjectives()[0] + ", wakeloss="+wakeloss+", secondsNeeded="+timeNeeded);
		
		
		
//		System.out.println("Finished algorithm with best fitness " + resultantPopulation.get(0).getObjectives()[0]);
	}
    
    
    
    public static Solution runDSOnce(Problem problem, Area area, Shape bounds, double x2, double y2, WakeModel fitfun, int populationSize, int lambda, int maxEvaluations, 
    		boolean printEveryEvaluation, boolean printSeveralOutputsPerDSRun, boolean displayGUI,
    		double directionSigma, double distanceMu, double distanceSigma,
    		int nearestNeighbours, double probability, double reversalProbability, Rectangle2D[] infeasibleZones) {
    	OpenWindMutation mutation = new OpenWindMutation(probability, nearestNeighbours, 
              bounds, area, x2, y2, windCode.WakeModel.R*8, /*directionSigma, distanceMu, distanceSigma,*/ reversalProbability);
    	
    	
//    	VectorBasedMovementMutation mutation = new VectorBasedMovementMutation(probability, nearestNeighbours, 
//                bounds, area, x2, y2, windCode.WakeModel.R*8, /*directionSigma, distanceMu, distanceSigma,*/ reversalProbability);
//        VectorBasedMovementMutation mutation = new VectorBasedMovementMutation(probability, nearestNeighbours, 
//                bounds, area, x2, y2, windCode.WakeModel.R*8, directionSigma, distanceMu, distanceSigma, reversalProbability);
		WindTurbineAlgorithm algorithm = new WindTurbineAlgorithm(problem, bounds, windCode.WakeModel.R*8, mutation, maxEvaluations, populationSize, 
				printEveryEvaluation, printSeveralOutputsPerDSRun, displayGUI, infeasibleZones);

    	SolutionSet resultantPopulation = algorithm.execute();
		return resultantPopulation.get(0);
    }
    
    
    
    public static Solution runDSSeveralTimes(Problem problem, Area area, Shape bounds, double x2, double y2, WakeModel fitfun, int mu, int lambda, int maxEvaluations,
            boolean printEveryEvaluation, boolean printSeveralOutputsPerDSRun, boolean printSeveralOutputsPerDSRepetition, boolean displayGUI,
            int repetitions, double directionSigma, double distanceMu, double distanceSigma,
            int nearestNeighbours, double probability, double reversalProbability, Rectangle2D[] infeasibleZones
//            int workerThreads
            ) {
        Solution bestSolution = null;
        Solution[] solutions = new Solution[repetitions];

        for (int rep = 0; rep<repetitions; rep++) {

            if (printSeveralOutputsPerDSRepetition) {
                System.out.print("$repetition:"+rep);
                System.out.println(" [" + DateUtils.now()+"]");
            }

            long startTime, stopTime;
            startTime = System.currentTimeMillis();
            Solution temp = runDSOnce(problem, area, bounds, x2, y2, fitfun, mu, lambda, maxEvaluations, 
            		printEveryEvaluation, printSeveralOutputsPerDSRun, displayGUI, 
            		directionSigma, distanceMu, distanceSigma,
            		nearestNeighbours, probability, reversalProbability, infeasibleZones);//, workerThreads);
            stopTime = System.currentTimeMillis();

            solutions[rep]=temp;

            // remember: CMA minimizes!
            if (bestSolution == null || (temp.getObjectives()[0] < bestSolution.getObjectives()[0])) {
                bestSolution = temp;
            }
            if (printSeveralOutputsPerDSRepetition)
                System.out.println("$repetition:"+rep+
                        ", newBest:"+1d/bestSolution.getObjectives()[0]+ ", bestHere:"+1d/temp.getObjectives()[0] +
                        " (time:+"+ (stopTime-startTime)/1000d +")"
                        );
        }

        if (printSeveralOutputsPerDSRepetition) {
            System.out.println("Best Solution found:");
            double[] energyPerTurbine = fitfun.energyPerTurbine(bestSolution.getDecisionVariableAsDoubleArray(), false);
            WakeModel.printSolution(bestSolution, energyPerTurbine);
        }

        //calculate mean:
        double mean = 0;
        for (Solution i : solutions) {
            mean += 1d/i.getObjectives()[0];
        }
        mean = mean/repetitions;
        double stdev = getStdDev(solutions);
        if (printSeveralOutputsPerDSRepetition)
                System.out.println(">> mean:"+mean+" stdev:"+stdev);

        return bestSolution;
    }

    
    
    
    
//    
//    public static Solution runDSOnce(Area area, WakeModel fitfun, int mu, int lambda, int generations,
//            boolean printSeveralOutputsPerDSRun, double initialStandardDeviation,
//            int workerThreads) {
//        long startTime2, stopTime2;
//        boolean printEvaluationTimes = !true;
////        area.printSeveralOutputs = printSeveralOutputsPerDSRun;
//
//        // new a CMA-ES and set some initial values
////        CMAEvolutionStrategy cma = new CMAEvolutionStrategy();
//        //cma.readProperties(); // read options, see file CMAEvolutionStrategy.properties
//        //cma.parameters.setCcov(0);
////        cma.parameters.setPopulationSize(lambda);
////        cma.parameters.setMu(mu);
////        cma.parameters.setPopulationSize(lambda);
//        /* Note: it looks like cma initally samples solutions until
//         * the initial populations contains feasible search points only.
//         * Thus, using 9 turbines on the 500m-radiusFarm "practically" never
//         * goes past the initialization.
//         */
//
////        cma.setDimension(2*areal.numberOfTurbines); // overwrite some loaded properties
//        //cma.setInitialX(0.0); // in each dimension, also setTypicalX can be used
//
//        double[] startingPoint = area.startingSolution();
//        double wakelossFreeFitness = fitfun.computeWakelossFreeFitness(startingPoint);
//        System.out.println("initial="+1d/fitfun.valueOf(startingPoint, true)+", wakelossfree="+1d/wakelossFreeFitness);
//        double[] energyPerTurbine = fitfun.energyPerTurbine(startingPoint, false);
//        WakeModel.printSolution(startingPoint, energyPerTurbine);
////        WakeModel.printSolution(startingPoint);
//
//
////        cma.setInitialX(startingPoint);
//
//        //cma.setInitialStandardDeviation(0.2); // also a mandatory setting
////        cma.setInitialStandardDeviation(initialStandardDeviation); // also a mandatory setting
//        //cma.setInitialStandardDeviation(fitfun.farmRadius/2); // reasonable stdev for initialization?
////        cma.options.stopFitness = 1e-19;       // optional setting, originally 1e-9;
//        
//
//        // initialize cma and get fitness array to fill in later
//        double[] fitness = cma.init();  // new double[cma.parameters.getPopulationSize()];
//
//        // initial output to files
////        cma.writeToDefaultFilesHeaders(0); // 0 == overwrites old files
//
//        // iteration loop
////        while (cma.stopConditions.getNumber() == 0) {
//        long startTime, stopTime;
//        int countResamplesPerRun = 0;
//        int countResamplesPerGeneration = 0;
//        while (cma.getCountIter() <= generations) {
//
//            startTime = System.currentTimeMillis();
//            // break if no improvement for long time
//            if (cma.getCountIter()>0 && (cma.getCountEval()-cma.getBestSolution().getEvaluationNumber())>lambda*500 ) {
//                System.out.println("break: no improvement for evals="+lambda*500);
//                break;
//            }
//
//            // two lines of code for more frequent output of the EA
//            if (printSeveralOutputsPerCMARun) {
//                if (cma.getCountIter() % 5 ==1 ) System.out.print("["+cma.getCountIter()+":"+1d/cma.getBestFunctionValue()+"]");
//                if (cma.getCountIter() % 25 ==0 && cma.getCountIter()>0) System.out.println();
//                if (cma.getCountIter() % 50 ==0 && cma.getCountIter()>0) {
//                    System.out.println("best placement found so far");
//                    energyPerTurbine = fitfun.energyPerTurbine(cma.getBestSolution().getX(), false);
//                    WakeModel.printSolution(cma.getBestSolution(), energyPerTurbine);
////                    WakeModel.printSolution(cma.getBestSolution());
//                }
//            }
//            if (printSeveralOutputsPerCMARun) {
//                System.out.println("[generation:" + cma.getCountIter() +
//                        ", bestSoFar: "+(1.0/cma.getBestFunctionValue())+
//                        ", distanceToWakelossFree:" + ((1.0/cma.getBestFunctionValue()) - (1.0/wakelossFreeFitness)) +
//                        ", evaluationNumber:" + cma.getBestEvaluationNumber() +
//                        "]");
//                System.out.println("[generation:" + cma.getCountIter() +
//                        ", bestRecent:"+(1.0/cma.getBestRecentFunctionValue())+
//                        ", distanceToWakelossFree:" + ((1.0/cma.getBestRecentFunctionValue()) - (1.0/wakelossFreeFitness)) +
//                        ", evaluationNumber:" + cma.getBestRecentSolution().getEvaluationNumber() +
//                        "]" );
//            }
//            //WakeModel.printSolution(cma.getBestSolution());
//
//            // fitness function can be modified on-the-fly (for problem 'Simple')
//            // (can be used to introduce penalty terms)
//            // if (cma.getCountIter() % 1000 == 0) fitfun.border = Math.random();
//
//            // --- core iteration step ---
//            //sample population
////            countResamplesPerGeneration = 0;
////            double[][] pop = cma.samplePopulation(areal); // get a new population of solutions
////            for (int i = 0; i < pop.length; ++i) {    // for each candidate solution i
////                // a simple way to handle constraints that define a convex feasible domain
////                // (like box constraints, i.e. variable boundaries) via "blind re-sampling"
////                // assumes that the feasible domain is convex, the optimum is
////                int countResamples = 0;
////
////                //while (!fitfun.isFeasible(pop[i], areal)) //   not located on (or very close to) the domain boundary,
////                
////        //pop[i] = areal.fixSolution(pop[i]); //possibly fix the newly sampled solution
////        //WakeModel.printSolution(pop[i]);
////                while (!fitfun.isFeasible(pop[i], areal)) //   not located on (or very close to) the domain boundary,
////                {
////                    pop[i] = cma.resampleSingle(i);    //   initialX is feasible and initialStandardDeviations are
////                    pop[i] = areal.fixSolution(pop[i]); //possibly fix the newly sampled solution
////                    countResamples++;
////                    // start from the beginning when 100 resamples were necessary!
////                    if (countResamples>100) {
////                        pop[i] = areal.startingSolution();
////                        System.out.println("!");// sublimal hint that the EA is having problems as too many resamples are needed
////                        break;
////                    }
////                }                                                       //   sufficiently small to prevent quasi-infinite looping here
////                if (countResamples>0 && false) System.out.println("resampled: "+countResamples);
////                // compute fitness/objective value
//////                System.out.println(fitness[i]);
////            //    fitness[i] = fitfun.valueOf(pop[i], printEvaluationTimes); // fitfun.valueOf() is to be minimized
//////                System.out.println(fitness[i]);
////                countResamplesPerRun += countResamples;
////                countResamplesPerGeneration += countResamples;
////            }
////
//            // make sure that areal.startingSolution() is contained at least once in the beginning
//            // actually: due to (X,Y)-ES it is not sure whether this helps at all
////            if (cma.getCountIter()==0 || cma.getCountIter()==1) {
////                pop[0] = areal.startingSolution();
////            }
////
//            //overwrite old fitness values
////            for (int i = 0; i < pop.length; ++i) {    // for each candidate solution i
////                fitness[i] = 0d;
////            }
////
////            
////            startTime2 = System.currentTimeMillis();
////
////            //now: determine fitnesses
//////            if (workerThreads == 1) {
//////                //old school: 1 by 1
//////                for (int i = 0; i < pop.length; ++i) {    // for each candidate solution i
//////    //                System.out.println("before"+fitness[i]);
//////                    fitness[i] = fitfun.valueOf(pop[i], printEvaluationTimes); // fitfun.valueOf() is to be minimized
//////    //                System.out.println("after "+fitness[i]);
//////                }
//////            } else {
//////                //parallelized :)
//////                
//////                ParallelizedEvaluation.parallelizeEvaluation(workerThreads, fitfun, pop, fitness, printEvaluationTimes);
//////                stopTime2 = System.currentTimeMillis();
//////                if (true) System.out.println("EndOfGeneration:"+(stopTime2-startTime2)/1000d+"s"+"("+Thread.currentThread().activeCount()+")");
//////            }
////
////            cma.updateDistribution(fitness);         // pass fitness array to update search distribution
////
////            stopTime2 = System.currentTimeMillis();
////            if (true) System.out.println("EndUpdateDistribution:"+(stopTime2-startTime2)/1000d+"s"+"("+Thread.currentThread().activeCount()+")");
////
//            // --- end core iteration step ---
////
//            // output to files and console
////            if (!true) {
////                cma.writeToDefaultFiles();
////                int outmod = 15;
////                if (cma.getCountIter() % (15 * outmod) == 1) {
////                    cma.printlnAnnotation(); // might write file as well
////                }
////                if (cma.getCountIter() % outmod == 1) {
////                    cma.println();
////                }
////            }
//
//            stopTime = System.currentTimeMillis();
//            if (printSeveralOutputsPerCMARun) System.out.println("seconds needed for this generation:"+(stopTime-startTime)/1000d);
//            if (printSeveralOutputsPerCMARun) System.out.println("countResamplesPerGeneration:"+countResamplesPerGeneration);
//            if (printSeveralOutputsPerCMARun) System.out.println("countResamplesPerRun:"+countResamplesPerRun);
//
//
//
//            if (cma.getCountIter() % 100==0) {
//            System.out.println(cma.sigma);
//            print2DArray(cma.C); }
//        }
//
//        if (printSeveralOutputsPerCMARun) System.out.println("["+(cma.getCountIter()-1)+":"+1d/cma.getBestFunctionValue()+"]");
//
//
//        // evaluate mean value as it is the best estimator for the optimum
////        cma.setFitnessOfMeanX(fitfun.valueOf(cma.getMeanX(), false)); // updates the best ever solution
//
//
//        //hide a lot of output
////        if (false) {
////            // final output
////            cma.writeToDefaultFiles(1);
////            cma.println();
////            cma.println("Terminated due to");
////            for (String s : cma.stopConditions.getMessages()) {
////                cma.println("  " + s);
////            }
////            cma.println("best function value " + 1d/cma.getBestFunctionValue()
////                    + " at evaluation " + cma.getBestEvaluationNumber());
////
////            // we might return cma.getBestSolution() or cma.getBestX()
////        }
//        //System.out.println("["+cma.getCountIter()+":"+1d/cma.getBestFunctionValue()+"]");
////        if (printSeveralOutputsPerCMARun) System.out.println("countResamplesPerRun:"+countResamplesPerRun);
//        return cma.getBestSolution();
//    }
//    
//    
    
    
    
    
    
    
    
    // own implementation to check whether Statistics.getStdDev() works correctly.
    // result: it does.
    public static double getStdDev(Solution[] input) {
        double result = 0;
        double mean = 0;
        double stdev = 0;

        // calculate mean
        for (int i = 0; i < input.length; i++) {
            mean += 1d/input[i].getObjectives()[0];
            //mean += input[i].doubleValue();
        }
        mean = mean/input.length;

        // calculate stdev
        for (int i = 0; i < input.length; i++) {
            stdev = stdev + Math.pow( 1d/input[i].getObjectives()[0] - mean, 2);
        }
        stdev = Math.sqrt(stdev/(input.length-1));

        //System.out.println(" mean:"+mean+" stdev:"+stdev);
        return stdev;

    }

}
