package directSpread;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;

import javax.swing.JFrame;
import directSpread.utility.Solution;
import directSpread.utility.Variable;
import java.awt.Dimension;


public class WindTurbineGUI extends JFrame{

	private Solution solution;
	private double minimumDistance;
	private Shape bounds;
	private double scale;
	private double maxScale = 20;
	private double turbineWidth = 5;
	private double turbineHeight = 5;

	private Shape[] infeasibleZones;

        int width = 800;
        int height = 800;
	
	/**
	 * Instantiate and display the representation GUI. Nodes are represented as
	 *  a black dot, minimum distances are represented by half-sized circles
	 *  around the nodes. Using this reduced form of minimum distance shows the 
	 *  rings unable to cross, rather than the nodes unable to cross a ring.
	 *  Slightly misleading, but effectively the same and neater!
	 * 
	 * @param minimumDistance  The minimum distance between the nodes.
	 * @param bounds  The bounding box for the nodes to lie within.
	 * @param infeasibleZones 
	 */
	public WindTurbineGUI(double minimumDistance, Shape bounds, int width, int height, Rectangle2D[] infeasibleZones){
		this.minimumDistance = minimumDistance;
		AffineTransform transformer = new AffineTransform();

		this.width = width;
		this.height = height;
		
//		double idealWidth = 4*getToolkit().getScreenSize().getWidth()/5;
//		double idealHeight = 4*getToolkit().getScreenSize().getHeight()/5;
		double idealWidth = 4*width/5;
		double idealHeight = 4*height/5;
		double currentBoundsWidth = bounds.getBounds().getWidth();
		double currentBoundsHeight = bounds.getBounds().getHeight();
		double scaleX = idealWidth/currentBoundsWidth;
		double scaleY = idealHeight/currentBoundsHeight;
		scale = (scaleX < scaleY) ? scaleX : scaleY;
		if (scale > maxScale){
			scale = maxScale;
		}
		
		transformer.scale(scale,scale);
		this.bounds = transformer.createTransformedShape(bounds);
		
		this.infeasibleZones = new Shape[infeasibleZones.length];
		for (int i = 0; i < infeasibleZones.length; i++){
			if (infeasibleZones[i] == null) break;
			double y = bounds.getBounds().getHeight() - infeasibleZones[i].getY() - infeasibleZones[i].getHeight();			
			
			Rectangle2D tempRectangle = new Rectangle2D.Double(infeasibleZones[i].getX(), y, infeasibleZones[i].getWidth(), infeasibleZones[i].getHeight());
			this.infeasibleZones[i] = transformer.createTransformedShape(tempRectangle);
		}

				
//		this.turbineWidth *= scale;
//		this.turbineHeight *= scale;
                this.turbineWidth = minimumDistance/50;
		this.turbineHeight = minimumDistance/50;
		this.minimumDistance*=scale;
                
		
		setDefaultCloseOperation(EXIT_ON_CLOSE);
//		setSize(getToolkit().getScreenSize());
		setSize(new Dimension(width, height));
		setVisible(true);
	}
	
	/**
	 * Pass in a new solution to display.
	 * 
	 * @param solution  The solution intended for display.
	 */
	public void updateSolution(Solution solution){
		this.solution = solution;
		repaint();
	}
	
	/**
	 * Paints the bounding shape and nodes to the window.
	 * Nodes are black dots, the bounding shape and the minimum distance
	 *  circles are blue.
	 * 
	 * @param g  The specified Graphics window
	 */
	public void paint(Graphics g){
		super.paint(g);
		
		Graphics2D g2d = (Graphics2D) g;
		int originX = width/2 - bounds.getBounds().width/2;
		int originY = height/2 - bounds.getBounds().height/2;
//		int originX = getToolkit().getScreenSize().width/2 - bounds.getBounds().width/2;
//		int originY = getToolkit().getScreenSize().height/2 - bounds.getBounds().height/2;
		g2d.translate(originX, originY);
		
		g2d.setColor(Color.BLUE);
		g2d.draw(bounds);
		for (int i = 0; i < infeasibleZones.length; i++){
			if (infeasibleZones[i] == null) break;
//			double x = infeasibleZones[i].getX()*scale;
//			double y = infeasibleZones[i].getY()*scale;
//			double height = infeasibleZones[i].getHeight()*scale;
//			double width = infeasibleZones[i].getWidth()*scale;
			
			//AffineTransform transformer = new AffineTransform();
			//transformer.scale(scale,scale);
			g2d.setColor(Color.CYAN);
	//		g2d.fill(new Rectangle2D.Double(x, y, width, height) );
			g2d.fill(infeasibleZones[i]);

		}
		
		
		//g2d.translate(0, -bounds.getBounds().getHeight());
		if (solution == null) return;
		
		Variable[] variables = solution.getDecisionVariables();
		

		g2d.setColor(Color.BLACK);
		
		for (int i = 0; i < variables.length/2; i++){
			double x = variables[2*i].getValue() * scale;
			double y = bounds.getBounds().height-variables[2*i+1].getValue() * scale;

			g2d.fill(new Ellipse2D.Double(x-turbineWidth/2, y-turbineHeight/2, turbineWidth, turbineHeight) );
			g2d.setColor(Color.BLUE);
			g2d.draw(new Ellipse2D.Double(x-minimumDistance/2, y-minimumDistance/2, minimumDistance, minimumDistance) );
			g2d.setColor(Color.BLACK);

		}
		
	}
}
