package directSpread.utility;

import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.util.Arrays;

import windCode.*;

public class WindTurbineProblem implements Problem, Serializable {

	private int numNodes;
	private double minimumDistance;
	private Shape bounds;
    private WakeModel wm;
    private Area area;
	private double initialDistanceMu;
	private double initialDistanceSigma;
	private double initialDirectionSigma;
	
	private double distanceMuIncrease;
	private double distanceSigmaIncrease;
	private double distanceMuDecrease;
	private double distanceSigmaDecrease;
	private double distanceSigmaMax;
	private double distanceSigmaMin;
	private double distanceMuMax;
	private double distanceMuMin;
		
	public WindTurbineProblem(int numNodes, double minimumDistance, Shape bounds,
			Area area, int wakeModelScenario, String realism,
			double initialDistanceMu, double initialDistanceSigma,
			double initialDirectionSigma, double distanceMuIncrease,
			double distanceSigmaIncrease, double distanceMuDecrease,
			double distanceSigmaDecrease, double distanceSigmaMax,
			double distanceSigmaMin, double distanceMuMax, double distanceMuMin){
		
		this.numNodes = numNodes;
		this.minimumDistance = minimumDistance;
		this.bounds = bounds;
		
            this.wm = new WakeModel(wakeModelScenario, realism);
            this.area = area;
//            areal = new CircularArea(turbs, windFarmRadius);
            
        this.initialDistanceMu = initialDistanceMu;
        this.initialDistanceSigma = initialDistanceSigma;
        this.initialDirectionSigma = initialDirectionSigma;
        
        this.distanceMuIncrease = distanceMuIncrease;
        this.distanceSigmaIncrease = distanceSigmaIncrease;
        this.distanceMuDecrease = distanceMuDecrease;
        this.distanceSigmaDecrease = distanceSigmaDecrease;
        this.distanceSigmaMax = distanceSigmaMax;
        this.distanceSigmaMin = distanceSigmaMin;
        this.distanceMuMax = distanceMuMax;
        this.distanceMuMin = distanceMuMin;
	}


        // evaluate layout using wakeModel
	public void evaluate(Solution newSolution) {

		Variable[] variables = newSolution.getDecisionVariables();

                // transfer Variable[] to double[]
                double[] coords = new double[variables.length];
                for (int i = 0; i<coords.length; i++) {
                    coords[i] = variables[i].getValue();
                }

                newSolution.setObjective(0, wm.valueOf(coords, false));
//                newSolution.setObjective(0, 1d/wm.valueOf(coords, false));
	}

	/**
	 * DOES NOT HANDLE ELLIPSES PROPERLY - treats as a rectangle
	 */
	public Variable[] createVariables() {

		double[] coordinates = area.startingSolution();

                // transfer double[] to Variable[]
		Variable[] vars = new Variable[coordinates.length];
		for (int i = 0; i < coordinates.length; i++){
			vars[i] = new Variable(coordinates[i]);
		}


		return vars;
	}
	
	public WakeModel getWakeModel() {
		return this.wm;
	}
	
	public Area getArea() {
		return this.area;
	}


	@Override
	public double[] getInitialDistanceMus() {
		double[] array = new double[numNodes];
		Arrays.fill(array, initialDistanceMu);
		return array;
	}


	@Override
	public double[] getInitialDistanceSigmas() {
		double[] array = new double[numNodes];
		Arrays.fill(array, initialDistanceSigma);
		return array;
	}


	@Override
	public double[] getInitialDirectionSigmas() {
		double[] array = new double[numNodes];
		Arrays.fill(array, initialDirectionSigma);
		return array;
	}


	public double getDistanceMuIncrease() {
		return distanceMuIncrease;
	}

	public double getDistanceSigmaIncrease() {
		return distanceSigmaIncrease;
	}

	public double getDistanceMuDecrease() {
		return distanceMuDecrease;
	}

	public double getDistanceSigmaDecrease() {
		return distanceSigmaDecrease;
	}

	public double getDistanceSigmaMax() {
		return distanceSigmaMax;
	}

	public double getDistanceSigmaMin() {
		return distanceSigmaMin;
	}

	public double getDistanceMuMax() {
		return distanceMuMax;
	}

	public double getDistanceMuMin() {
		return distanceMuMin;
	}

	

	
//	//A greater distance is a better fitness.
//	public void evaluate2(Solution newSolution) {
//
//		Variable[] variables = newSolution.getDecisionVariables();
//		double sum = 0;
//
//		for (int node1 = 0; node1 < variables.length/2; node1++){
//			for (int node2 = node1+1; node2 < variables.length/2; node2++){
//				double xDiff = variables[2*node2].getValue() - variables[2*node1].getValue();
//				double yDiff = variables[2*node2+1].getValue() - variables[2*node1+1].getValue();
//
//				sum += 1/Math.sqrt(xDiff*xDiff + yDiff*yDiff);
//			}
//		}
//		newSolution.setObjective(0,sum);
//
//	}
//
//	/**
//	 * DOES NOT HANDLE ELLIPSES PROPERLY - treats as a rectangle
//	 */
//	public Variable[] createVariables() {
//
//		Rectangle actualBounds = bounds.getBounds();
//		double[] coordinates = new double[numNodes*2];
//
//		int pointer = 0;
//		double x = actualBounds.getX();
//		while (x <= actualBounds.width && pointer < coordinates.length){
//			double y = actualBounds.getY();
//			while (y <= actualBounds.height && pointer < coordinates.length){
//				coordinates[pointer++] = x;
//				coordinates[pointer++] = y;
//
//				y+=minimumDistance;
//			}
//			x+=minimumDistance;
//		}
//
//		if (pointer < coordinates.length){
//			System.err.println("Error in WindTurbineProblem.createVariables: Not enough room for nodes.");
//			System.exit(1);
//		}
//
//		Variable[] vars = new Variable[coordinates.length];
//		for (int i = 0; i < coordinates.length; i++){
//			vars[i] = new Variable(coordinates[i]);
//		}
//
//		return vars;
//	}

}