package windCode;

/**
 *
 * @author mwagner
 */
public class CircularArea extends Area {
    public double windFarmRadius = 0;

    public CircularArea(int numberOfTurbines, double windFarmRadius) {
        this.numberOfTurbines = numberOfTurbines;
        this.windFarmRadius = windFarmRadius;
    }

    public double[] fixSolution(double[] coords) {
        //System.out.println("WARNING: fixSolution(double[] coords) not implemented, returning 'coords'");
        boolean debugPrint = !true;

        // split coords s.t. x[] and y[]
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        // constraint 1: be within circle of radius R
        for (int i = 0; i < x.length; i++) {
            double oldDistanceFromCenter = Math.sqrt(x[i] * x[i] + y[i] * y[i]);
            if (oldDistanceFromCenter > this.windFarmRadius) {
                if (debugPrint) System.out.println("fixSolution: "+oldDistanceFromCenter+">"+this.windFarmRadius);
                //infeasible --> fix this
                double angle = Math.asin(y[i] / oldDistanceFromCenter);
                if (debugPrint) System.out.println("old: "+coords[i*2]+" "+coords[i*2+1]);
                coords[i*2] = (this.windFarmRadius) * Math.cos(angle); //x[i]
                coords[i*2+1] = (this.windFarmRadius) * Math.sin(angle); //y[i]
                if (debugPrint) System.out.println("new: "+coords[i*2]+" "+coords[i*2+1]);
            }
        }

        return coords;
    }

    public double[] startingSolution() {
        double[] result = new double[2*numberOfTurbines];

        //initialize non-intelligently, hard-coded for Kusik-paper
        double[] result1 = {0, 0};
        double[] result2 = {0, 250, 0, -250};
        //double[] result2 = {-75, 494, 100, -400};
        double[] result3 = {-250, 250, -250, -250, 250, -250};
        double[] result4 = {-250, 250, -250, -250, 250, -250, 250, 250};
        double[] result5 = {-330, 330, -330, 0, -330, -330, 0, 330, 0, 0};
        double[] result6 = {-330, 330, -330, 0, -330, -330, 0, 330, 0, 0, 0, -330};
        double[] result7 = {-330, 330, -330, 0, -330, -330, 0, 330, 0, 0, 0, -330, 330, 330};
        double[] result8 = {-330, 330, -330, 0, -330, -330, 0, 330, 0, 0, 0, -330, 330, 330, 330, 0};
        double[] result9 = {-330, 330, -330, 0, -330, -330, 0, 330, 0, 0, 0, -330, 330, 330, 330, 0, 330, -330};
        double[] result21={-400, 800, 0, 800, 400, 800,
                -800, 400, -400, 400, 0, 400, 400, 400, 800, 400,
                -800,   0, -400,   0, 0,   0, 400,   0, 800,   0,
                -800,-400, -400,-400, 0,-400, 400,-400, 800,-400,
                           -400,-800, 0,-800, 400,-800}; // for a radius of 1200

        switch(numberOfTurbines) {
            case 1: result = result1; break;
            case 2: result = result2; break;
            case 3: result = result3; break;
            case 4: result = result4; break;
            case 5: result = result5; break;
            case 6: result = result6; break;
            case 7: result = result7; break;
            case 8: result = result8; break;
            case 9: result = result9; break;
            case 21: result = result21; break;
        }

        return result;
    }

    public boolean isFeasible(double[] coords) {
        // split coords s.t. x[] and y[]
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        // constraint 1: be within circle of radius R
        for (int i = 0; i < x.length; i++) {
            if (Math.sqrt(x[i] * x[i] + y[i] * y[i]) > this.windFarmRadius) {
                return false; //infeasible
            }
        }

        // constraint 2: turbines placements should adhere to safety distances: 4 rotor diameters
        for (int i = 0; i < x.length; i++) {
            for (int j = 0; j < x.length; j++) {
                if (i!=j) {
                    if ( Math.sqrt( Math.pow(x[i]-x[j], 2) + Math.pow(y[i]-y[j], 2) ) < 4*(2*WakeModel.R))
                        return false; //infeasible
                }
            }
        }
        return true;
    }

    public String toString() {
        String result = "CircularArea(turbines="+this.numberOfTurbines+",radius="+this.windFarmRadius;
        return result;
    }

    public double computeInitialStandardDeviation() {
            return 2d; // return a magic number
    }

}