package windCode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import deepCopy.*;

/**
 *
 * @author mwagner
 */
class ParallelizedEvaluationOLD {

    public static void parallelizeEvaluation(int workerThreads, WakeModel wakeModel, double[][] pop, double[] fitnesses, boolean printEvaluationTimes) {
        long startTime, stopTime;
        startTime = System.currentTimeMillis();

        int pollEveryMs = 10;

        int limit = workerThreads;// number of worker threads

        int overallIndex = 0;
        double[] o = fitnesses;

        boolean reducedPrintouts = false;
        boolean printALot = false;

        try {
            if (reducedPrintouts) {
                System.out.println(" active threads: " + Thread.activeCount());
            }

            //the giant loop to do everything in parallel...
            while (true) {
                //if (overallIndex == o.length) break;

                if (printALot) {
                    printDoubleArray(o);
                    System.out.println("threads running: " + Thread.activeCount());
                }

                //1. may I start new threads?
                if (Thread.activeCount() < 1 + limit) {

                    //2. now start new threads
                    //2.1. what is left?
//                    for (int i = overallIndex; i < o.length; i++) {
                    for (int i = overallIndex; overallIndex < o.length; i++) {

                        //if something had not yet been computed...
//                        if (o[i] == 0d) {
                            //...start thread
                            Thread temp = new Thread(new WorkerThread(wakeModel, pop, fitnesses, i, printEvaluationTimes));
                            temp.start();
//                            temp.sleep(100);
//                            System.out.println(Thread.activeCount());

                            //Thread.sleep(1000);

                            if (printALot) {
                                Thread[] ta = new Thread[Thread.activeCount()];
                                Thread.enumerate(ta);
                                System.out.print("> ids: ");
                                for (int t = 0; t < ta.length; t++) {
                                    if (ta[t] != null) {
                                        System.out.print(ta[t].getName() + " ");
                                    }
                                }
                                System.out.println();
                            }

                            if (printALot) {
                                System.out.println("starting thread... total:" + Thread.activeCount() + " last started on i=" + i);
                            }
                            overallIndex++;
//                        }
                        // enough threads running? --> break 'start new threads'
                        if (Thread.activeCount() >= 1 + limit) {
                            break;
                        }
                    }
                }

                //wait because the threads now have to work...
                if (printALot) {
                    System.out.println("------------------------------ waiting");
                }
                //Thread.sleep(100);
                //synchronized (o) {o.wait();}

                //magic assumption on the underlying hardware, as "busy waiting" needs one core (if more than 4 cores present: don't care...)
                if (workerThreads>4) Thread.sleep(pollEveryMs);
//                Runtime r = Runtime.getRuntime();r.gc();


                //2.2 if nothing is left: break main loop
                boolean somethingLeft = false;
                for (int i = 0; i < o.length; i++) {
                    if (o[i] == 0d) {
                        somethingLeft = true;
                        if (printALot) {
                            System.out.println("something is left");
                        }
                        break;
                    }
                }
                if (!somethingLeft) {
                    if (printALot) {
                        System.out.println("nothing is left:");
                        printDoubleArray(o);
                    }
                    break;
                }

                if (printALot) {
                    System.out.println();
                    System.out.println("------------------------------ waiting STOPPED");
                    System.out.println(" active threads: " + Thread.activeCount());
                }

            }//END main loop

            stopTime = System.currentTimeMillis();

            if (reducedPrintouts) {
                System.out.println("---> time needed: " + (stopTime - startTime) / 1000d);
                System.out.println("all threads done");
                printDoubleArray(o);
                System.out.println(" number of worker threads: " + limit);
            }

        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }


    public static void printDoubleArray(double[] da) {
        String result = "[";
        for (int i = 0; i < da.length - 1; i++) {
            result += da[i] + ",";
        }
        result += da[da.length - 1] + "]";
        System.out.println(result);
    }
}


class WorkerThread implements Runnable, Callable {

    public WakeModel fitfun;
    public double[][] pop;
    public double[] fitnesses;
    public int index = 0;
    public boolean printEvaluationTimes;

//    public ParallelizedEvaluation(WakeModel fitfun, double[][] pop, double[] fitnesses, int index, boolean printEvaluationTimes) {
    public WorkerThread(WakeModel fitfun, double[][] pop, double[] fitnesses, int index, boolean printEvaluationTimes) {
        this.fitfun = fitfun;
        this.pop = pop;
        this.fitnesses = fitnesses;
        this.index = index;
        this.printEvaluationTimes = printEvaluationTimes;
    }

    public void run() {
//do heavy computation
//        System.out.println("run() before "+fitnesses[index]);
        long startTime, stopTime;
        startTime = System.currentTimeMillis();
        fitnesses[index] = fitfun.valueOf(pop[index], printEvaluationTimes);
        stopTime = System.currentTimeMillis();
        if (!true) System.out.println(index+":"+(stopTime-startTime)/1000d+"s"+"("+Thread.currentThread().activeCount()+")");
//        System.gc();
//            o[index] = Math.random();
//        System.out.println("run() after  "+fitnesses[index]);
    }

    public Double call() {
        long startTime, stopTime;
        startTime = System.currentTimeMillis();
        double result =  fitfun.valueOf(pop[index], printEvaluationTimes);
        stopTime = System.currentTimeMillis();
        if (!true) System.out.println(index+":"+(stopTime-startTime)/1000d+"s"+"("+Thread.currentThread().activeCount()+")");
        return result;
    }
}


class ParallelizedEvaluation {

    public static void parallelizeEvaluation(int workerThreads, WakeModel wakeModel, double[][] pop, double[] fitnesses, boolean printEvaluationTimes) {
        long startTime, stopTime;
        startTime = System.currentTimeMillis();
        
        ExecutorService exec = Executors.newFixedThreadPool(workerThreads);

//        Collection<WorkerThread> tasks = new ArrayList<WorkerThread>();
        for (int i=0; i<fitnesses.length; i++) {
            WakeModel wakeModelCopy = (WakeModel) DeepCopy.copy(wakeModel);
            double[][] popCopy = (double[][]) DeepCopy.copy(pop);
            int iCopy = (Integer)DeepCopy.copy(i);
            boolean printEvaluationTimesCopy = (Boolean) DeepCopy.copy(printEvaluationTimes);

            exec.execute(new WorkerThread(wakeModelCopy, popCopy, fitnesses, iCopy, printEvaluationTimesCopy));
//            exec.execute(new WorkerThread(wakeModelCopy, popCopy, fitnesses, i, printEvaluationTimes));
//            tasks.add(new WorkerThread(wakeModel, pop, fitnesses, i, printEvaluationTimes));
        }

//  List<Future<Double>> results = exec.invokeAll(tasks)invokeAll(tasks);
        exec.shutdown();
        
        try {
            exec.awaitTermination(30l, TimeUnit.MINUTES);
        } catch (InterruptedException ex) {
            Logger.getLogger(ParallelizedEvaluation.class.getName()).log(Level.SEVERE, null, ex);
        }
        stopTime = System.currentTimeMillis();
        if (true) System.out.println("EndOfGenerations:"+(stopTime-startTime)/1000d+"s"+"("+Thread.currentThread().activeCount()+")");

        
//        try {
//            if (false) Thread.sleep(1000 * 10);
//            System.gc();
//        } catch (InterruptedException ex) {
//            Logger.getLogger(ParallelizedEvaluation.class.getName()).log(Level.SEVERE, null, ex);
//        }
  }


    public static void printDoubleArray(double[] da) {
        String result = "[";
        for (int i = 0; i < da.length - 1; i++) {
            result += da[i] + ",";
        }
        result += da[da.length - 1] + "]";
        System.out.println(result);
    }
}