package windCode;

import java.awt.geom.Rectangle2D;

/**
 *
 * @author mwagner
 */
public class RectangularArea extends Area {
    public double windFarmWidth = 0;
    public double windFarmHeight = 0;
    public Rectangle2D[] infeasibleZones;

    public RectangularArea(int numberOfTurbines, double windFarmWidth, double windFarmHeight, Rectangle2D[] infeasibleZones) {
    	this(numberOfTurbines, windFarmWidth, windFarmHeight);
        this.infeasibleZones = infeasibleZones;
    }
    
    public RectangularArea(int numberOfTurbines, double windFarmWidth, double windFarmHeight) {
        this.numberOfTurbines = numberOfTurbines;
        this.windFarmWidth = windFarmWidth;
        this.windFarmHeight = windFarmHeight;
    }

    public Rectangle2D[] getInfeasibleZones(){
    	return infeasibleZones;
    }

    public double[] fixSolution(double[] coords) {
        boolean printFixes = !true;

        // constraint 1: be within area of windFarm
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        if (printFixes) {
            System.out.println("FIXED: START");
            WakeModel.printSolution(coords);
        }

        for (int i = 0; i < x.length; i++) {
            if (x[i]>=this.windFarmWidth) {
                //return false; //infeasible
                if (printFixes) System.out.print("fix "+coords[i*2]+" to ");
                coords[i*2] = this.windFarmWidth;
                x[i] = this.windFarmWidth;
                if (printFixes) System.out.println(coords[i*2]);
            }
            if (y[i]>=this.windFarmHeight) {
                //return false; //infeasible
                if (printFixes) System.out.print("fix "+coords[i*2+1]+" to ");
                coords[i*2+1] = this.windFarmHeight;
                y[i] = this.windFarmHeight;
                if (printFixes) System.out.println(coords[i*2+1]);
            }
            if (x[i]<=0) {
                //return false; //infeasible
                if (printFixes) System.out.print("fix "+coords[i*2]+" to ");
                coords[i*2] = 0;
                x[i] = 0;
                if (printFixes) System.out.println(coords[i*2]);
            }
            if (y[i]<=0) {
                //return false; //infeasible
                if (printFixes) System.out.print("fix "+coords[i*2+1]+" to ");
                coords[i*2+1] = 0;
                y[i] = 0;
                if (printFixes) System.out.println(coords[i*2+1]);
            }
        }

        if (printFixes) {
            WakeModel.printSolution(coords);
            System.out.println("FIXED: END");
        }

        return coords;
    }

    //place turbines on borders
    public double[] startingSolution() {

        if (numberOfTurbines==2) return new double[]{0, 0, this.windFarmWidth, this.windFarmHeight};

        double[] result = new double[2*numberOfTurbines];
        boolean debugPrint = !true; //boolean printSeveralOutputs = true;

        //intial check: is feasible solution possible?
        //System.out.println(this.windFarmWidth/(8.0*WakeModel.R));
        int turbinesInWidthPossible = 1+ (int) Math.floor(this.windFarmWidth/(8.0*WakeModel.R)); //+1 possible
        int turbinesInHeightPossible = 1+ (int) Math.floor(this.windFarmHeight/(8.0*WakeModel.R));
        int tempI = turbinesInHeightPossible*turbinesInWidthPossible;
        
        if (printSeveralOutputs) System.out.println("turbines possible: "+turbinesInWidthPossible+"*"+turbinesInHeightPossible+"="+tempI+"(needed:"+this.numberOfTurbines+")");
        
        //Attempt to fit all turbines into the area. If there are infeasible zones, allow room for more turbines.
        int turbineOffset = 0;
        int turbineOffsetIncrement = 10;
        int turbinesToPlace;
		do{
			turbinesToPlace = numberOfTurbines;
		    int redundantTurbines = numberOfTurbines+turbineOffset;   
		    if (redundantTurbines > tempI && redundantTurbines < tempI + turbineOffsetIncrement){
		    	redundantTurbines = tempI;
		    }
		//	        
		//        if (this.numberOfTurbines > tempI) {
		//            System.out.println("Greedy check for feasibility failed (tight row-wise packing)");
		//            System.out.println("no feasible solution for "+ this.numberOfTurbines +" turbines possible! max turbines possible:"+tempI);
		//            System.out.println("[exiting...]");
		//            System.exit(0);
		//        }
		
		    
		    if (redundantTurbines > tempI) {
		        System.out.println("Greedy check for feasibility failed (tight row-wise packing)");
		        System.out.println("no feasible solution for "+ this.numberOfTurbines +" turbines possible in the space provided.");
		        System.out.println("[exiting...]");
		        System.exit(0);
		    }
    
		        //now construct initial solution...
		
		        //1. evenly reduce the number of turbines per row/column (row:width, column:height)
		        int turbinesPerRow = turbinesInWidthPossible;
		        int turbinesPerColumn = turbinesInHeightPossible;
		        boolean firstIsBigger = false;
		        int x, y;
		        if (turbinesPerRow>turbinesPerColumn) {
		            firstIsBigger = true;
		            x = turbinesPerRow;
		            y = turbinesPerColumn;
		        } else {
		            firstIsBigger = false;
		            x = turbinesPerColumn;
		            y = turbinesPerRow;
		        }
		        while(true) {
		            int tempX = x;
		            for (int i = x; i> tempX*((y-1d)/(double)y); i--) {
		                if (debugPrint) System.out.println(tempX*((y-1d)/(double)y));
		                if (debugPrint) System.out.println(x*((y-1d)/(double)y));
		                if (debugPrint) System.out.println("X "+x+"*"+y+"="+x*y+"("+this.numberOfTurbines+")");
		                x--;
		                if (x*y<redundantTurbines) {
		                    x++;
		                    if (debugPrint) System.out.println("break x");
		                    break;
		                }
		            }
		
		            // now reduce smaller 'dimension'
		            if (debugPrint) System.out.println("Y "+x+"*"+y+"="+x*y+"("+this.numberOfTurbines+")");
		            y--;
		            if (x*y<redundantTurbines) {
		                y++;
		                if (debugPrint) System.out.println("break y");
		                break;
		            }
		        }
		        //reassign x, y
		        if (firstIsBigger) {
		            turbinesPerRow = x;
		            turbinesPerColumn = y;
		        } else {
		            turbinesPerRow = y;
		            turbinesPerColumn = x;
		        }
		        if (printSeveralOutputs) System.out.println("distribution: "+turbinesPerRow+"*"+turbinesPerColumn+"="+turbinesPerRow*turbinesPerColumn+"("+this.numberOfTurbines+")");
		
		        
		        // now construct the start solution
		        //1. compute avg distances between turbines:
		
		        System.out.println(WakeModel.R*8 + " " + this.windFarmWidth/((double)turbinesPerRow-1d)+ " " + this.windFarmHeight/((double)turbinesPerColumn-1d));
		
		        /*spacing: the next 2 lines: not tight, but not too generous spacing*/
		//        double avgDistanceInRow = ( WakeModel.R*8 + this.windFarmWidth/((double)turbinesPerRow-1d) )/2d;
		//        double avgDistanceInColumn = ( WakeModel.R*8 + this.windFarmHeight/((double)turbinesPerColumn-1d) )/2d;
		
		        /*spacing: the next 2 lines: maximal generous spacing*/
		        double avgDistanceInRow = this.windFarmWidth/((double)turbinesPerRow-1d);
		        double avgDistanceInColumn = this.windFarmHeight/((double)turbinesPerColumn-1d);
		
		        /*spacing: the next 2 lines: minimal spacing*/
		//        double avgDistanceInRow = WakeModel.R*8;
		//        double avgDistanceInColumn = WakeModel.R*8;
		
		        if (printSeveralOutputs) System.out.println("avgdistances: inRow:"+avgDistanceInRow + " inColumn:"+avgDistanceInColumn + "(min:8R="+8*WakeModel.R+")");
		        double firstTurbineX = 0;//(this.windFarmWidth-(avgDistanceInRow*(turbinesPerRow-1d)))/2d;
		        double firstTurbineY = 0;//(this.windFarmHeight-(avgDistanceInColumn*(turbinesPerColumn-1d)))/2d;
		        
                        
                        //2. distribute the turbines...
		        
                        /* addition: improved spacing, by making holes from time to time */
                        int emptyFields = turbinesPerRow * turbinesPerColumn - turbinesToPlace;
//                        double total = turbinesPerRow * turbinesPerColumn;
//                        double emptyPercent = (double)emptyFields/((double)turbinesPerRow * turbinesPerColumn);
                        int leaveEmptyDistance = 1; // this sets how far apart the "holes" (missing turbines) should be when evenly spread out in the wind farm
                        int startEmptyField = (turbinesPerRow * turbinesPerColumn) - (leaveEmptyDistance * (emptyFields+1)) ;
                        if (emptyFields!=0) {
                            leaveEmptyDistance = turbinesToPlace / (emptyFields);
                        }
                        
                        startEmptyField = startEmptyField/2;
//                        startEmptyField = 0;
//                        int startEmptyField = turbinesToPlace / (emptyFields + 1);
//                        int lastTurbineLeftOut = 0;
                        int gridPosition = 0;
                        int skipped = 0;
                        System.out.println(" emptyFields="+emptyFields+" leaveEmptyDistance="+leaveEmptyDistance);
                        
		    	int currentPtr = 0;
		
		        for (int i=0; i<turbinesPerRow; i++) {
		        	        	
		            for (int j=0; j<turbinesPerColumn && turbinesToPlace>0 ;j++) { //checks whether there is actually a turbine left to be placed
                                
                                if ((gridPosition+startEmptyField) % leaveEmptyDistance == 0)
                                  if (emptyFields!=0)
                                  if (gridPosition!=0) 
                                  if (skipped<emptyFields) 
                                  if (leaveEmptyDistance>=2)
                                  {
                                    System.out.println(" "+gridPosition+" skipped due to improved grid spacing ("+turbinesToPlace+" to go)");
                                    gridPosition++;
                                    skipped++;
                                    continue; //skip the next position
                                } 
		            	
                                // this is the newly calculated turbine position
		            	double currentX = firstTurbineX + i*avgDistanceInRow;
		            	double currentY = firstTurbineY + j*avgDistanceInColumn;
		            	boolean feasible = true;
                                
                                // check if the newly computed turbine position is in an infeasible area
                                if (infeasibleZones==null) infeasibleZones = new Rectangle2D[]{new Rectangle2D.Double(0, 1, 0, 0)};
                                for (Rectangle2D infeasibleZone : infeasibleZones) {
//                                    if (infeasibleZone == null) {break;}// unreachable code?
                                    if (infeasibleZone.contains(currentX, currentY)) {
                                        System.out.println(" "+gridPosition+" skipped due to infeasibility ("+turbinesToPlace+" to go)");
                                        skipped++;
                                        feasible = false;
                                        break;
                                    }
                                }
		            	
		            	if (feasible){
		            		result[currentPtr++] = currentX;
		            		result[currentPtr++] = currentY;
		            		turbinesToPlace--;
                                        System.out.println(" one placed at gridPosition="+gridPosition+" ("+currentX+","+currentY+") ("+turbinesToPlace+" to go)");
                                        gridPosition++;
		            	}
		            }
		        }
		        
		        turbineOffset+=turbineOffsetIncrement;
		}while(turbinesToPlace > 0);

        return result;
    }


//    public double[] startingSolutionNOTONBORDER() {
//        double[] result = new double[2*numberOfTurbines];
//        boolean debugPrint = true; //boolean printSeveralOutputs = true;
//
//        //intial check: is feasible solution possible?
//        //System.out.println(this.windFarmWidth/(8.0*WakeModel.R));
//        int turbinesInWidthPossible = 1+ (int) Math.floor(this.windFarmWidth/(8.0*WakeModel.R)); //+1 possible
//        int turbinesInHeightPossible = 1+ (int) Math.floor(this.windFarmHeight/(8.0*WakeModel.R));
//        int tempI = turbinesInHeightPossible*turbinesInWidthPossible;
//        if (printSeveralOutputs) System.out.println("turbines possible: "+turbinesInWidthPossible+"*"+turbinesInHeightPossible+"="+tempI+"(needed:"+this.numberOfTurbines+")");
//        if (this.numberOfTurbines > tempI) {
//            System.out.println("Greedy check for feasibility failed (tight row-wise packing)");
//            System.out.println("no feasible solution for "+ this.numberOfTurbines +" turbines possible! max turbines possible:"+tempI);
//            System.out.println("[exiting...]");
//            System.exit(0);
//        }
//
//        //now construct initial solution...
//
//        //1. evenly reduce the number of turbines per row/column (row:width, column:height)
//        int turbinesPerRow = turbinesInWidthPossible;
//        int turbinesPerColumn = turbinesInHeightPossible;
//        boolean firstIsBigger = false;
//        int x, y;
//        if (turbinesPerRow>turbinesPerColumn) {
//            firstIsBigger = true;
//            x = turbinesPerRow;
//            y = turbinesPerColumn;
//        } else {
//            firstIsBigger = false;
//            x = turbinesPerColumn;
//            y = turbinesPerRow;
//        }
//        while(true) {
//            int tempX = x;
//            for (int i = x; i> tempX*((y-1d)/(double)y); i--) {
//                if (debugPrint) System.out.println(tempX*((y-1d)/(double)y));
//                if (debugPrint) System.out.println(x*((y-1d)/(double)y));
//                if (debugPrint) System.out.println("X "+x+"*"+y+"="+x*y+"("+this.numberOfTurbines+")");
//                x--;
//                if (x*y<this.numberOfTurbines) {
//                    x++;
//                    if (debugPrint) System.out.println("break x");
//                    break;
//                }
//            }
//
//            // now reduce smaller 'dimension'
//            if (debugPrint) System.out.println("Y "+x+"*"+y+"="+x*y+"("+this.numberOfTurbines+")");
//            y--;
//            if (x*y<this.numberOfTurbines) {
//                y++;
//                if (debugPrint) System.out.println("break y");
//                break;
//            }
//        }
//        //reassign x, y
//        if (firstIsBigger) {
//            turbinesPerRow = x;
//            turbinesPerColumn = y;
//        } else {
//            turbinesPerRow = y;
//            turbinesPerColumn = x;
//        }
//        if (printSeveralOutputs) System.out.println("distribution: "+turbinesPerRow+"*"+turbinesPerColumn+"="+turbinesPerRow*turbinesPerColumn+"("+this.numberOfTurbines+")");
//
//        // now construct the start solution
//        //1. compute avg distances between turbines:
//
//        System.out.println(WakeModel.R*8 + " " + this.windFarmWidth/((double)turbinesPerRow-1d));
//        double avgDistanceInRow = ( WakeModel.R*8 + this.windFarmWidth/((double)turbinesPerRow-1d) )/2d;
//        double avgDistanceInColumn = ( WakeModel.R*8 + this.windFarmHeight/((double)turbinesPerColumn-1d) )/2d;
////        double avgDistanceInRow = WakeModel.R*8;
////        double avgDistanceInColumn = WakeModel.R*8;
//        if (printSeveralOutputs) System.out.println("avgdistances: inRow:"+avgDistanceInRow + " inColumn:"+avgDistanceInColumn + "(min:8R="+8*WakeModel.R+")");
//        double firstTurbineX = (this.windFarmWidth-(avgDistanceInRow*(turbinesPerRow-1d)))/2d;
//        double firstTurbineY = (this.windFarmHeight-(avgDistanceInColumn*(turbinesPerColumn-1d)))/2d;
//        //2. distribute the turbines...
//        int turbinesToPlace = this.numberOfTurbines;
//        for (int i=0; i<turbinesPerRow; i++) {
//            for (int j=0; j<turbinesPerColumn && turbinesToPlace>0 ;j++) { //checks whether there is actually a turbine left to be placed
//                //x
//                result[2*(i*turbinesPerColumn+j)] = firstTurbineX + i*avgDistanceInRow;
//                //y
//                result[2*(i*turbinesPerColumn+j)+1] = firstTurbineY + j*avgDistanceInRow;
//                turbinesToPlace--;
//            }
//        }
//
//        return result;
//    }


    // the following: new isFeasible(...) with correction of coordinates
    public boolean isFeasible(double[] coords) {
        // split coords s.t. x[] and y[]
        //int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        boolean printFixes = !true;

        if (printFixes) System.out.println("FIXED: START");
        if (printFixes) WakeModel.printSolution(coords);

        // constraint 1: be within area of windFarm
//        for (int i = 0; i < x.length; i++) {
//            if (x[i]>=this.windFarmWidth) {
//                //return false; //infeasible
//                if (printFixes) System.out.print("fix "+coords[i*2]+" to ");
//                coords[i*2] = this.windFarmWidth;
//                x[i] = this.windFarmWidth;
//                if (printFixes) System.out.println(coords[i*2]);
//            }
//            if (y[i]>=this.windFarmHeight) {
//                //return false; //infeasible
//                if (printFixes) System.out.print("fix "+coords[i*2+1]+" to ");
//                coords[i*2+1] = this.windFarmHeight;
//                y[i] = this.windFarmHeight;
//                if (printFixes) System.out.println(coords[i*2+1]);
//            }
//            if (x[i]<=0) {
//                //return false; //infeasible
//                if (printFixes) System.out.print("fix "+coords[i*2]+" to ");
//                coords[i*2] = 0;
//                x[i] = 0;
//                if (printFixes) System.out.println(coords[i*2]);
//            }
//            if (y[i]<=0) {
//                //return false; //infeasible
//                if (printFixes) System.out.print("fix "+coords[i*2+1]+" to ");
//                coords[i*2+1] = 0;
//                y[i] = 0;
//                if (printFixes) System.out.println(coords[i*2+1]);
//            }
//        }

        if (printFixes) WakeModel.printSolution(coords);
        if (printFixes) System.out.println("FIXED: END");

        // constraint 2: turbines placements should adhere to safety distances: 4 rotor diameters
        for (int i = 0; i < x.length; i++) {
            for (int j = 0; j < x.length; j++) {
                if (i!=j) {
                    if ( Math.sqrt( Math.pow(x[i]-x[j], 2) + Math.pow(y[i]-y[j], 2) ) < 4*(2*WakeModel.R))
                        return false; //infeasible
                }
            }
        }
        return true;
    }


    // the following: old isFeasible(...) without correction of coordinates
    public boolean isFeasibleWITHOUTCORRECTION(double[] coords) {
        // split coords s.t. x[] and y[]
        //int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        // constraint 1: be within area of windFarm
        for (int i = 0; i < x.length; i++) {
            if (x[i]>=this.windFarmWidth) return false; //infeasible
            if (y[i]>=this.windFarmHeight) return false; //infeasible
            if (x[i]<=0) return false; //infeasible
            if (y[i]<=0) return false; //infeasible
        }

        // constraint 2: turbines placements should adhere to safety distances: 4 rotor diameters
        for (int i = 0; i < x.length; i++) {
            for (int j = 0; j < x.length; j++) {
                if (i!=j) {
                    if ( Math.sqrt( Math.pow(x[i]-x[j], 2) + Math.pow(y[i]-y[j], 2) ) < 4*(2*WakeModel.R))
                        return false; //infeasible
                }
            }
        }
        return true;
    }


    public String toString() {
        String result = "RectangularArea(turbines="+this.numberOfTurbines+",width="+this.windFarmWidth+",height="+this.windFarmHeight+")";
        return result;
    }


    public double computeInitialStandardDeviation() {
        int numberOfTurbines = this.numberOfTurbines;
        double width = this.windFarmWidth;
        double height = this.windFarmHeight;
        double result = 0;
        boolean debugPrint = false;

        int turbinesInWidthPossible = 1+ (int) Math.floor(width/(8.0*WakeModel.R)); //+1 possible
        int turbinesInHeightPossible = 1+ (int) Math.floor(height/(8.0*WakeModel.R));
        int tempI = turbinesInHeightPossible*turbinesInWidthPossible;
        if (debugPrint) System.out.println("[turbines max possible: "+turbinesInWidthPossible+"*"+turbinesInHeightPossible+"="+tempI+"]\n");
        double avgDistanceInRow = ( WakeModel.R*8 + width/((double)turbinesInWidthPossible-1d) )/2d;
        double avgDistanceInColumn = ( WakeModel.R*8 + height/((double)turbinesInHeightPossible-1d) )/2d;
        if (debugPrint) System.out.println("[avgdistances then: inRow:"+avgDistanceInRow + " inColumn:"+avgDistanceInColumn + "(min:8R="+8*WakeModel.R+")]");

                //1. evenly reduce the number of turbines per row/column (row:width, column:height)
            int turbinesPerRow = turbinesInWidthPossible;
            int turbinesPerColumn = turbinesInHeightPossible;
            boolean firstIsBigger = false;
            int x, y;
            if (turbinesPerRow>turbinesPerColumn) {
                firstIsBigger = true;
                x = turbinesPerRow;
                y = turbinesPerColumn;
            } else {
                firstIsBigger = false;
                x = turbinesPerColumn;
                y = turbinesPerRow;
            }
            while(true) {
                int tempX = x;
                for (int i = x; i> tempX*((y-1d)/(double)y); i--) {
                    if (debugPrint) System.out.println(x*((y-1d)/(double)y));
                    if (debugPrint) System.out.println("X "+x+"*"+y+"="+x*y+"("+numberOfTurbines+")");
                    x--;
                    if (x*y<numberOfTurbines) {
                        x++;
                        break;
                    }
                }

                // now reduce smaller 'dimension'
                if (debugPrint) System.out.println("Y "+x+"*"+y+"="+x*y+"("+numberOfTurbines+")");
                y--;
                if (x*y<numberOfTurbines) {
                    y++;
                    break;
                }
            }
            //reassign x, y
            if (firstIsBigger) {
                turbinesPerRow = x;
                turbinesPerColumn = y;
            } else {
                turbinesPerRow = y;
                turbinesPerColumn = x;
            }
            if (debugPrint) System.out.println("distribution: "+turbinesPerRow+"*"+turbinesPerColumn+"="+turbinesPerRow*turbinesPerColumn+"("+numberOfTurbines+")");

            // now construct the start solution
            //1. compute avg distances between turbines:
            avgDistanceInRow = ( WakeModel.R*8 + width/((double)turbinesPerRow-1d) )/2d;
            avgDistanceInColumn = ( WakeModel.R*8 + height/((double)turbinesPerColumn-1d) )/2d;
            if (debugPrint) System.out.println("avgdistances: inRow:"+avgDistanceInRow + " inColumn:"+avgDistanceInColumn + "(min:8R="+8*WakeModel.R+")\n");


            //divisor: magic number: 8-->98,5% are within +-4stdevs, 6-->97.2% are within +-3stdevs
            result = Math.min(avgDistanceInRow-8*WakeModel.R, avgDistanceInColumn-8*WakeModel.R)/32d;
            result = Math.max(result, 1.0); //just prevent stdev from being 0.0 at the beginning
            return result;
        }
}