package windCode.backups;

import cma.CMASolution;
import java.text.DecimalFormat;

import directSpread.utility.*;
import java.util.Arrays;
import windCode.Area;
import windCode.RectangularArea;

/**
 *
 * @author mwagner
 */

//class WakeModel implements IObjectiveFunction{
public class WakeModelFasterfarmlayoutbased implements java.io.Serializable{
    public String wakeModel = "";
    //parameters and scenario from "Kusiak 2010 Design of wind farm layout for maximum wind energy capture"

    public static double R = 38.5; // rotor radius

    public double cutInSpeed = 3.5;
    public double ratedSpeed = 14;
    public double[] windSpeeds = {cutInSpeed, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0,
                                       9.5, 10.0, 10.5, 11.0, 11.5, 12.0, 12.5, 13.0, 13.5, ratedSpeed}; // v_0 = v_cutin, v_last = v_ratedSpeed

    public double ratedPower = 1500; // in kilowatts

    // wind distribution
    public double[] windks = new double[0];
    public double[] windcs = new double[0];
    public double[] windthetas = {      15,  30,  45,  60,  75,  90, 105, 120, 135, 150, 165, 180,
                                       195, 210, 225, 240, 255, 270, 285, 300, 315, 330, 345, 360};
    public double[] windthetasMinusOne = {0,  15,  30,  45,  60,  75,  90, 105, 120, 135, 150, 165, 180,
                                             195, 210, 225, 240, 255, 270, 285, 300, 315, 330, 345};
    //public double windSectorSize = 15; // most likely not needed?
    public double[] windSectorOmegasMinusOne = new double[0];

    // for linear power curve function
    public double lambda = 140.86;
    public double eta = -500;

    //public double hubHeight = 80;
    public double CT = 0.8; // thrust coefficient
    public double kappa = 0.075; // spreading konstant for land cases

    //public double farmRadius = 0; // wind farm radius, will be set in the scenario
    Area area;

    public WakeModelFasterfarmlayoutbased(int scenario, String wakeModel) {

        if (scenario==1){
            double[] windks1 = {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                            2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};
            this.windks = windks1;
            double[] windcs1 =  {13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
                13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13}; // wind Weibull scale parameter (is the same in this scenario for all wind directions)
            this.windcs = windcs1;
            double[] windSectorOmegasMinusOne1 = {0,    0.01, 0.01, 0.01, 0.01, 0.2,  0.6,  0.01, 0.01, 0.01, 0.01, 0.01,
                                                  0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0};
            this.windSectorOmegasMinusOne = windSectorOmegasMinusOne1;

        } else if (scenario==2){
            double[] windks2 = {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                            2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};
            this.windks = windks2;
            double[] windcs2 =  {7, 5, 5, 5, 5, 4, 5, 6, 7, 7, 8, 9.5,
                10, 8.5, 8.5, 6.5, 4.6, 2.6, 8, 5, 6.4, 5.2, 4.5, 3.9};
            this.windcs = windcs2;
            double[] windSectorOmegasMinusOne2 = {0.0002, 0.008, 0.0227, 0.0242, 0.0225, 0.0339,  0.0423,  0.029, 0.0617, 0.0813, 0.0994, 0.1394,
                                                  0.1839, 0.1115, 0.0765, 0.008, 0.0051, 0.0019, 0.0012, 0.001, 0.0017, 0.0031, 0.0097, 0.0317};
            this.windSectorOmegasMinusOne = windSectorOmegasMinusOne2;

        } else if (scenario==3){
            double[] windks3 = {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                            2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};
            this.windks = windks3;
            double[] windcs3 =  {5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
                5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5};
            this.windcs = windcs3;
            double[] windSectorOmegasMinusOne3 = {0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0,
                                                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
            this.windSectorOmegasMinusOne = windSectorOmegasMinusOne3;

        }

        this.wakeModel = wakeModel;
    }



    // get energy per turbine (very similar to valueOf/2!)
    public double[] energyPerTurbine(double[] coords, boolean printEvaluationTimes, FarmLayout f, int update) {

        double[] result = new double[coords.length/2];

        long startTime, stopTime;
        startTime = System.currentTimeMillis();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

//        double result = 0;
        for (int i=0; i<numberOfTurbines; i++) {

            double temp = 0;

            if (this.wakeModel.compareTo("o")==0) {
                temp = energyOfTurbineOPTIMAL(x, y, i);
            } else if (this.wakeModel.compareTo("w")==0) {
                temp = energyOfTurbineWAKELOSS(x, y, i, f, update);
            }

            result[i] = temp;
            
            if (f!=null)
                f.energyPerTurbine[i] = temp;
            
        }

        stopTime = System.currentTimeMillis();
        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");

        return result;
    }
    
    
    
    // get energy for a turbine (very similar to energyPerTurbine!)
    public double energyForOneTurbine(double[] coords, int turbineNumber, boolean printEvaluationTimes, FarmLayout f, int update) {

        double result;

        long startTime, stopTime;
        startTime = System.currentTimeMillis();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        int i = turbineNumber;

        double temp = 0;

        if (this.wakeModel.compareTo("o")==0) {
            temp = energyOfTurbineOPTIMAL(x, y, i);
        } else if (this.wakeModel.compareTo("w")==0) {
            temp = energyOfTurbineWAKELOSS(x, y, i, f, update);
        }

        result = temp;
        

        stopTime = System.currentTimeMillis();
        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");

        return result;
    }
    
    
    
    
    
    // the fitness function
    public double valueOf(boolean printEvaluationTimes, FarmLayout f, int update) {

        double[] coords = f.coordinates;
        
        long startTime, stopTime;
        startTime = System.currentTimeMillis();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        double result = 0;
        for (int i=0; i<numberOfTurbines; i++) {

            double temp = 0;

            if (this.wakeModel.compareTo("o")==0) {
                temp = energyOfTurbineOPTIMAL(x, y, i);
            } else if (this.wakeModel.compareTo("w")==0) {
                temp = energyOfTurbineWAKELOSS(x, y, i, f, update);
            }

            result += temp;
            
            if (f!=null) 
                f.energyPerTurbine[i] = temp;
        }

        stopTime = System.currentTimeMillis();
        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");

// mediocre idea to penalize infeasible solutions. "mediocre" because of the (,)-strategy all generated solutions may be infeasible
//        if (isFeasible == false) {
//            result = result/2d;
//            System.out.println("i");
//        }

        if (f!=null)
            f.totalEnergy = result;
        
        return 1.0/result;
    }
    
    
    



    // the fitness function
    public double valueOf(double[] coords, boolean printEvaluationTimes, FarmLayout f, int update) {

//        boolean isFeasible = true;
//        // split coords s.t. x[] and y[]
//        //int numberOfTurbines = coords.length/2;
//        int numberOfTurbines = coords.length/2;
//        double[] x = new double[numberOfTurbines];
//        double[] y = new double[numberOfTurbines];
//        for (int i=0; i<numberOfTurbines; i++) {
//            x[i] = coords[i*2];
//            y[i] = coords[i*2+1];
//        }
//
//        // constraint 1: be within area of windFarm
//        for (int i = 0; i < x.length; i++) {
//            if (x[i]>=10000 || x[i]<=0) isFeasible = false; //infeasible
//            if (y[i]>=20000 || y[i]<=0) isFeasible = false; //infeasible
//        }
//
//        // constraint 2: turbines placements should adhere to safety distances: 4 rotor diameters
//        for (int i = 0; i < x.length; i++) {
//            for (int j = 0; j < x.length; j++) {
//                if (i!=j) {
//                    if ( Math.sqrt( Math.pow(x[i]-x[j], 2) + Math.pow(y[i]-y[j], 2) ) < 4*(2*WakeModel.R))
//                        isFeasible = false; //infeasible
//                }
//            }
//        }


        long startTime, stopTime;
        startTime = System.currentTimeMillis();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        double result = 0;
        for (int i=0; i<numberOfTurbines; i++) {

            double temp = 0;

            if (this.wakeModel.compareTo("o")==0) {
                temp = energyOfTurbineOPTIMAL(x, y, i);
            } else if (this.wakeModel.compareTo("w")==0) {
                temp = energyOfTurbineWAKELOSS(x, y, i, f, update);
            }

            result += temp;
            
            if (f!=null) 
                f.energyPerTurbine[i] = temp;
        }

        stopTime = System.currentTimeMillis();
        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");

// mediocre idea to penalize infeasible solutions. "mediocre" because of the (,)-strategy all generated solutions may be infeasible
//        if (isFeasible == false) {
//            result = result/2d;
//            System.out.println("i");
//        }

        if (f!=null)
            f.totalEnergy = result;
        
        return 1.0/result;
    }

    public double computeWakelossFreeFitness(double[] coords) {
        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }
        double result = 0;
        for (int i=0; i<numberOfTurbines; i++) {
            double temp = 0;
            temp = energyOfTurbineOPTIMAL(x, y, i);
            result += temp;
        }
        return 1.0/result;
    }

    // formula 22 OPTIMAL, WITHOUT WAKELOSS
    public double energyOfTurbineOPTIMAL(double[] x, double[] y, int i) {

        double result = 0;

        double temp = 0;

        double firstSummand = 0;
        double secondSummand = 0;
        double thirdSummand = 0;

        //FIRST
        temp = 0;
        for (int j = 1; j<this.windSpeeds.length; j++) {
            double avgSpeed = (this.windSpeeds[j-1] + this.windSpeeds[j]) / 2.0;
            double temp1 = 0;
            for (int l = 0; l < this.windthetas.length; l++) {
                temp1 += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                        this.windSectorOmegasMinusOne[l] *
                        (
                        Math.pow(Math.E, -(
                            Math.pow( this.windSpeeds[j-1]/c(thetaAvg(l)),
                                      k(thetaAvg(l))) )) -
                        Math.pow(Math.E, -(
                            Math.pow( this.windSpeeds[j  ]/c(thetaAvg(l)),
                                      k(thetaAvg(l))) ))
                        );
            }
            temp += avgSpeed * temp1;
        }

        firstSummand = lambda * temp;
        //FIRST

        //SECOND
        temp = 0;
        for (int l = 0; l<this.windthetas.length; l++) {
            temp += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                     this.windSectorOmegasMinusOne[l] *
                     Math.pow(
                         Math.E,
                         - Math.pow( (this.ratedSpeed/c(thetaAvg(l))),
                                     k(thetaAvg(l)))  );
        }
        secondSummand = this.ratedPower * temp;
        //SECOND

        //THIRD
        temp = 0;

        for (int l = 0; l<this.windthetas.length; l++) {
            temp += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                    this.windSectorOmegasMinusOne[l] *
                    (
                        Math.pow(Math.E, -(
                            Math.pow( this.cutInSpeed/c(thetaAvg(l)),
                                      k(thetaAvg(l))) )) -
                        Math.pow(Math.E, -(
                            Math.pow( this.ratedSpeed/c(thetaAvg(l)),
                                      k(thetaAvg(l))) ))
                        )
                    ;
        }

        thirdSummand = this.eta * temp;
        //THIRD

        result = firstSummand + secondSummand + thirdSummand;
        //System.out.println(result +" "+firstSummand+ " "+secondSummand+ " "+thirdSummand);
        return result;
    }

    // formula 18
    public double energyOfTurbineWAKELOSS(double[] x, double[] y, int i, FarmLayout f, int update) {

        double result = 0;

        double temp = 0;

        double firstSummand = 0;
        double secondSummand = 0;
        double thirdSummand = 0;

        //optimization to save computation time:
        double[] tempCiThetaAvgL = new double[this.windthetas.length];
        double[] tempKThetaAvgL = new double[this.windthetas.length];
        for (int l = 0; l < this.windthetas.length; l++) {
            tempCiThetaAvgL[l] = ci(thetaAvg(l), x, y, i, f, update);
            tempKThetaAvgL[l] = k(thetaAvg(l));
        }


        //FIRST
        temp = 0;
        for (int j = 1; j<this.windSpeeds.length; j++) {
            double avgSpeed = (this.windSpeeds[j-1] + this.windSpeeds[j]) / 2.0;
            double temp1 = 0;
            for (int l = 0; l < this.windthetas.length; l++) {
                temp1 += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                        this.windSectorOmegasMinusOne[l] *
                        (
                        Math.pow(Math.E, -(
                            Math.pow( this.windSpeeds[j-1]/tempCiThetaAvgL[l],
                                      tempKThetaAvgL[l]) )) -
                        Math.pow(Math.E, -(
                            Math.pow( this.windSpeeds[j  ]/tempCiThetaAvgL[l],
                                      tempKThetaAvgL[l]) ))
                        );
            }
            temp += avgSpeed * temp1;
        }

        firstSummand = lambda * temp;
        //FIRST

        //SECOND
        temp = 0;
        for (int l = 0; l<this.windthetas.length; l++) {
            temp += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                     this.windSectorOmegasMinusOne[l] *
                     Math.pow(
                         Math.E,
                         - Math.pow( this.ratedSpeed/tempCiThetaAvgL[l],
                                     tempKThetaAvgL[l])  );
        }
        secondSummand = this.ratedPower * temp;
        //SECOND

        //THIRD
        temp = 0;

        for (int l = 0; l<this.windthetas.length; l++) {
            temp += (this.windthetas[l]-this.windthetasMinusOne[l]) *
                    this.windSectorOmegasMinusOne[l] *
                    (
                        Math.pow(Math.E, -(
                            Math.pow( this.cutInSpeed/tempCiThetaAvgL[l],
                                      tempKThetaAvgL[l]) )) -
                        Math.pow(Math.E, -(
                            Math.pow( this.ratedSpeed/tempCiThetaAvgL[l],
                                      tempKThetaAvgL[l]) ))
                        )
                    ;
        }

        thirdSummand = this.eta * temp;
        //THIRD

        //System.out.println(firstSummand+ " "+secondSummand+ " "+thirdSummand);
        result = firstSummand + secondSummand + thirdSummand;
        return result;
    }

    // helper function to compute (theta_l+theta_l-1)/2
    public double thetaAvg(int l) {
        double result = 0;
        result = (this.windthetas[l] + this.windthetasMinusOne[l]) / 2.0;
        return result;
    }

    // formula 15
    public double ci(double theta, double[] x, double[] y, int i, FarmLayout f, int update) {
        double result = 0;

        result = c(theta) * (1.0 - Vel_def(x, y, i, theta, f, update));

        return result;
    }
    
    
    public int getThetaIndex(double theta){
        //technical detail: which theta-bucket are we currently in?
        return (int)(theta/15);
    }
    
    
    
    public double Vel_def(double[] x, double[] y, int i, double theta, FarmLayout f, int update) {
        boolean debugPrint = !true;
        if (i == update) {
            if (debugPrint) System.out.println("### Vel_def: starting Vel_defCOMPLETE with i"+i+" update"+update);
            return Vel_defCOMPLETE(x, y, i, theta, f);
        }
       
        if (update!=-1) {
            if (debugPrint) System.out.println("### Vel_def: starting Vel_defUPDATE with i"+i+" update"+update);
            return Vel_defUPDATE(x, y, i, theta, f, update);
        }
        else {
            if (debugPrint) System.out.println("### Vel_def: starting Vel_defCOMPLETE with i"+i+" update"+update);
            return Vel_defCOMPLETE(x, y, i, theta, f);
        }
    }
        

    // formula 14, velocity deficit at turbine i for angle theta
    public double Vel_defCOMPLETE(double[] x, double[] y, int i, double theta, FarmLayout f) {

        boolean debugPrint = !true;
        
        double result = 0;
        double temp = 0;
        boolean farmLayoutGiven = true;
        
        
        //technical detail: which theta-bucket are we currently in?
        int tempIndex = getThetaIndex(theta);
        
        
        if (debugPrint) {
            System.out.println("Vel_defCOMPLETE"+i+" "+Arrays.toString(f.thetaVeldefiVector[tempIndex])+ " "+theta);
        }

        for (int j = 0; j<x.length; j++) {
            if (j!=i) {
                double betaIJ = betaIJ(x, y, i, j, theta);
                double alpha = Math.abs( Math.atan(this.kappa)/Math.PI*180d );
                if (betaIJ < alpha) {
                    temp = Vel_def(x, y, i, j, theta);
//                    temp = Vel_def(x, y, i, j, theta) * Vel_def(x, y, i, j, theta);
                    
                    
                    // update matrix for later usage
                    if (f!=null) 
                        f.thetaVeldefijMatrix[tempIndex][i][j] = temp;
                    else {
//                        if (farmLayoutGiven) System.out.println("--- WARNING --- Vel_def: no Farmlayout given ---");
                        farmLayoutGiven = false; // prevent subsequent printouts of this warning
                    }
                    
                    
                    result += temp * temp;
                }
            }
        }

        result = Math.sqrt(result);
        
        // update vector for later usage
        if (f!=null) 
            f.thetaVeldefiVector[tempIndex][i] = result;
        else {
//            if (farmLayoutGiven) System.out.println("--- WARNING --- Vel_def: no Farmlayout given ---");
            farmLayoutGiven = false; // prevent subsequent printouts of this warning
        }
        
        if (debugPrint) {
            System.out.println("Vel_defCOMPLETEx "+Arrays.toString(f.thetaVeldefiVector[tempIndex]));
        }
        
        return result;
    }
    
    
    
    
    public double Vel_defUPDATE(double[] x, double[] y, int i, double theta, FarmLayout f, int update) {
        int tempIndex = getThetaIndex(theta);

        boolean debugPrint = !true;
        
        if (debugPrint) System.out.println("Vel_defUPDATE /// i"+i+" update"+update+" f.turbineMoved[update]="+f.turbineMoved[update]);
        
        double result = 0;
        double temp = 0;
        boolean farmLayoutGiven = true;
        int jj=0;
        
        int j = update;
//        for (int j = 0; j<x.length; j++) {
//        for (int j = 0; j<x.length; j++) {
            
            if (f.turbineMoved[j]) {
//                System.out.println("f.turbineMoved[update]="+f.turbineMoved[j]);
                jj = j;
//                if (debugPrint) System.out.println("theta"+theta+" tempIndex"+tempIndex);
                if (debugPrint) System.out.println("Vel_defUPDATE: i"+i+" j"+j+ " Vel_defi^old"+f.thetaVeldefiVector[tempIndex][i]+ " Vel_defij^old"+f.thetaVeldefijMatrix[tempIndex][i][j]);
            
                if (j!=i) {
                    double betaIJ = betaIJ(x, y, i, j, theta);
                    double alpha = Math.abs( Math.atan(this.kappa)/Math.PI*180d );
                    if (betaIJ < alpha) {
                        temp = Vel_def(x, y, i, j, theta);
//                        temp = Vel_def(x, y, i, j, theta) * Vel_def(x, y, i, j, theta);
//                        result += temp;

                        
                    }
                    
                    
                    result = Math.pow(f.thetaVeldefiVector[tempIndex][i],2) - 
                         Math.pow(f.thetaVeldefijMatrix[tempIndex][i][j],2) + 
                         temp * temp;

                    // update matrix for later usage
                        if (f!=null) 
                            f.thetaVeldefijMatrix[tempIndex][i][j] = temp;
                        else {
    //                        if (farmLayoutGiven) System.out.println("--- WARNING --- Vel_def: no Farmlayout given ---");
                            farmLayoutGiven = false; // prevent subsequent printouts of this warning
                        }
                    
//                } else {
//                    continue;
                    
                    result = Math.sqrt(result);
                }
//
            } else {
                System.out.println("!!! Vel_defUPDATE PROBLEM");
            }
//        }

        
        
        // update vector for later usage
        if (f!=null) 
            f.thetaVeldefiVector[tempIndex][i] = result;
        else {
//            if (farmLayoutGiven) System.out.println("--- WARNING --- Vel_def: no Farmlayout given ---");
            farmLayoutGiven = false; // prevent subsequent printouts of this warning
        }
        
        if (debugPrint) System.out.println("Vel_defUPDATE: i"+i+" j"+jj+ " Vel_defi^new"+f.thetaVeldefiVector[tempIndex][i]+ " Vel_defij^new"+f.thetaVeldefijMatrix[tempIndex][i][jj]);
        
        return result;
    }
    

    // formula 9
    public double betaIJ(double[] x, double[] y, int i, int j, double theta) {
        double result = 0;
        result = Math.acos(
                ( (x[i]-x[j])*Math.cos(theta*Math.PI/180d) + (y[i]-y[j])*Math.sin(theta*Math.PI/180d) + R/this.kappa )
                /
                (Math.sqrt(  Math.pow( (x[i]-x[j]+R/this.kappa*Math.cos(theta*Math.PI/180d)), 2) +
                             Math.pow( (y[i]-y[j]+R/this.kappa*Math.sin(theta*Math.PI/180d)), 2))
                 ))/Math.PI*180d;
        return result;
    }

    // formula 7
    public double Vel_def(double[] x, double[] y, int i, int j, double theta) {
        double result = 0;

        double a = 1 - Math.sqrt(1-this.CT);
        double b = this.kappa / R;

        result = a / Math.pow(1 + b * d(x, y, i, j, theta), 2);

        return result;
    }
    
    
    
    
    

    public double d(double[] x, double[] y, int i, int j, double theta) {
        double result = 0;

        //result = Math.abs( (x[i*2]-x[j*2]) * Math.cos(theta) + (x[i*2+1]*x[j*2+1])*Math.sin(theta) );
        result = Math.abs( (x[i]-x[j])*Math.cos(theta*Math.PI/180d) + (y[i]-y[j])*Math.sin(theta*Math.PI/180d) );
        return result;
    }

    // returns the wind Weibull scale parameter - dependend on theta
    public double c(double theta) {
        double result = 0;
        int index = -1;
        //lookup the index in the windthetas-table, and then return the corresponding windc
        for (int i = 0; i<this.windthetas.length; i++) {
            if (Math.abs( this.windthetas[i] - theta) <0.01 ) {
                System.out.println("function c(theta)!");
                index = i;
                break;
            }
        }

        if (index != -1) {
            // right theta + corresponding index were found
            result = this.windcs[index];
        } else {
            // take avg of two neighboring windthetas resp neighboring windcs
            for (int i = 0; i<this.windthetas.length; i++) {
                // when "windthetas[i]>theta" then the right intermediate index was found
                if (windthetas[i]>theta) {
                    int j = i-1;
                    if (j==-1) j = windcs.length-1;

                    // the following two lines make important performance differences!!!!!!!!!!!!!!!!!!
                    // "ideal" means: ideal according to Kusiak 2010 --> Kusiak's implementation/formula is buggy.
                    /* 1nd scenario
                     *  2 turbines -> ideal 28091.47, 1st line 28091.47, 2nd line 28091.47
                     *  3 turbines -> ideal 42137.21, 1st line 42137.21, 2nd line 42137.21
                     *  4 turbines -> ideal 56182.95, 1st line 56182.95, 2nd line 56182.95
                     *  5 turbines -> ideal 70228.69, 1st line 70228.69, 2nd line 70228.69
                     *  6 turbines -> ideal 84274.42, 1st line 84274.42, 2nd line 84274.42
                     *  --> no surprise that my numbers match, as 'c' is constant in this scenario
                     *      however, this is a good test of my implementation
                     */
                    /* 2nd scenario
                     *  2 turbines -> ideal 14631.37, 1st line 14226.72, 2nd line 14630.76  (one solution: 14630.60)
                     *  3 turbines -> ideal 21947.06, 1st line 21340.07, 2nd line 21946.14  (one solution: 21942.75)
                     *  4 turbines -> ideal 29262.74, 1st line 28453.43, 2nd line 29261.51  (one solution: 29252.87)
                     *  5 turbines -> ideal 36578.43, 1st line 35566.79, 2nd line 36576.89  (one solution: 36420.13)
                     *  6 turbines -> ideal 43894.11, 1st line 42680.15, 2nd line 43892.27  (one solution: 43729.67)
                     *  --> tiny deviations, maybe due to rounding errors in scenario table
                     *  --> rougly 0.5-1.5% improvement over Kusiak 2010
                     * furthermore: number of possible turbines: 7, ... (just checked whether feasible solutions found)
                     *              (confirmed via drawing: 9 turbines are possible to place using a "squared grid")
                     */
                    //result = (windcs[i] + windcs[j]) / 2.0; //that would be correct according to the official formula: the avg of two neighboring sectors.
                    result = windcs[i];

                    break;
                }
            }
        }

        //if (true) System.out.println("theta:"+theta+" c(theta):"+result);

        return result;
    }

    // returns the wind Weibull shape parameter - dependend on theta (CURRENTLY NOT DEPENDEND)
    public double k(double theta) {
//        double result = 0;
//        int index = 0;
//        //lookup the index in the windthetas-table, and then return the corresponding windk
//        for (int i = 0; i<this.windthetas.length; i++) {
//            if (Math.abs( this.windthetas[i] - theta) <0.01 ) {
//                index = i;
//                break;
//            }
//        }
//        result = this.windks[index];
//        return result;
        return 2;
    }

    // function that checks whether a current solution is feasible, dependent on the area
    public boolean isFeasible(double[] coords, Area areal) {
        return areal.isFeasible(coords);
    }

        // helper function to nicely print a solution
    public static void printSolution(CMASolution cs, double[] energyPerTurbine) {
        double[] coords = cs.getX();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        String s = "values: x, y, energyPerTurbine\n";
        for (int i = 0; i<x.length; i++) {
            s += (x[i]) + " " +(y[i]) + " " + energyPerTurbine[i];
            s += "\n";
        }
        System.out.println(s);
    }
    
    
    
    public static void printSolution(Solution cs, double[] energyPerTurbine) {
        double[] coords = cs.getDecisionVariableAsDoubleArray();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        String s = "values: x, y, energyPerTurbine\n";
        for (int i = 0; i<x.length; i++) {
            s += (x[i]) + " " +(y[i]) + " " + energyPerTurbine[i];
            s += "\n";
        }
        System.out.println(s);
    }

    public static void printSolution(double[] c, double[] energyPerTurbine) {
        double[] coords = c;

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

        String s = "values: x, y, energyPerTurbine\n";
        for (int i = 0; i<x.length; i++) {
            s += (x[i]) + " " +(y[i]) + " " + energyPerTurbine[i];
            s += "\n";
        }
        System.out.println(s);
    }


    // helper function to nicely print a solution
    public static void printSolution(CMASolution cs) {
        double[] coords = cs.getX();

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

//        DecimalFormat df = new DecimalFormat(" 000.000");
//        DecimalFormat dfneg = new DecimalFormat("000.000");
        String s = "";
        for (int i = 0; i<x.length-1; i++) {
//            s += "  "+(x[i]>0?df.format(x[i]):dfneg.format(x[i]))
//               + " " +(y[i]>0?df.format(y[i]):dfneg.format(y[i]));
            s += ""+x[i]
               + " " +y[i];
//            s += "  (distFromOrigin:"+Math.sqrt(x[i]*x[i] + y[i]*y[i])+")";
            s += "\n";
        }
//        s += "  "+(x[x.length-1]>0?df.format(x[x.length-1]):dfneg.format(x[x.length-1]))
//           + " " +(y[y.length-1]>0?df.format(y[y.length-1]):dfneg.format(y[y.length-1]));
        s += ""+(x[x.length-1])
           + " " +(y[y.length-1]);
//        s += "  (distFromOrigin:"+Math.sqrt(x[x.length-1]*x[x.length-1] + y[y.length-1]*y[y.length-1])+")";
        System.out.println(s);
    }


    // helper function to nicely print a solution
    public static void printSolution(double[] c) {
        double[] coords = c;

        // split coords s.t. x[] and y[]
        int numberOfTurbines = coords.length/2;
        double[] x = new double[numberOfTurbines];
        double[] y = new double[numberOfTurbines];
        for (int i=0; i<numberOfTurbines; i++) {
            x[i] = coords[i*2];
            y[i] = coords[i*2+1];
        }

//        DecimalFormat df = new DecimalFormat(" 000.000");
//        DecimalFormat dfneg = new DecimalFormat("000.000");
        String s = "";
        for (int i = 0; i<x.length-1; i++) {
//            s += "  "+(x[i]>0?df.format(x[i]):dfneg.format(x[i]))
//               + " " +(y[i]>0?df.format(y[i]):dfneg.format(y[i]));
            s += ""+(x[i])
               + " " +(y[i]);
//            s += "  (distFromOrigin:"+Math.sqrt(x[i]*x[i] + y[i]*y[i])+")";
            s += "\n";
        }
//        s += "  "+(x[x.length-1]>0?df.format(x[x.length-1]):dfneg.format(x[x.length-1]))
//           + " " +(y[y.length-1]>0?df.format(y[y.length-1]):dfneg.format(y[y.length-1]));
        s += ""+(x[x.length-1])
           + " " +(y[y.length-1]);
//        s += "  (distFromOrigin:"+Math.sqrt(x[x.length-1]*x[x.length-1] + y[y.length-1]*y[y.length-1])+")";
        System.out.println(s);
    }



    public static void main(String[] args) {
        
//        double[][][] t = new double[24][1000][1000];
//        t[2][3][4] = 3;
//        t[23][999][999] = 3;
        
        
        // the fitness function
//    public double valueOf(double[] coords, boolean printEvaluationTimes) {
        boolean printEvaluationTimes = true;
    double[] coords = { 1000, 2000, 3000, 4000};
    coords = config1;
    
    Area area = new RectangularArea(100, 20000, 10000);
    coords = area.startingSolution();
    
    

    FarmLayout f = new FarmLayout(coords);
    
    WakeModelFasterfarmlayoutbased w = new WakeModelFasterfarmlayoutbased(2,"w");
//
        long startTime, stopTime;
//        startTime = System.currentTimeMillis();
//
//        // split coords s.t. x[] and y[]
//        int numberOfTurbines = coords.length/2;
//        double[] x = new double[numberOfTurbines];
//        double[] y = new double[numberOfTurbines];
//        for (int i=0; i<numberOfTurbines; i++) {
//            x[i] = coords[i*2];
//            y[i] = coords[i*2+1];
//        }
//
//        double result = 0;
//        for (int i=0; i<numberOfTurbines; i++) {
//
//            double temp = 0;
//
////            if (this.wakeModel.compareTo("o")==0) {
////                temp = energyOfTurbineOPTIMAL(x, y, i);
////            } else if (this.wakeModel.compareTo("w")==0) {
//                temp = w.energyOfTurbineWAKELOSS(x, y, i, null);
////            }
//
//            result += temp;
//        }
//
//        stopTime = System.currentTimeMillis();
//        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");
        
// mediocre idea to penalize infeasible solutions. "mediocre" because of the (,)-strategy all generated solutions may be infeasible
//        if (isFeasible == false) {
//            result = result/2d;
//            System.out.println("i");
//        }

//        return 1.0/result;
//        System.out.println(result);
        
        
    
    // initial computation
        w.valueOf(printEvaluationTimes, f, -1);
        System.out.print(f.totalEnergy+" ");
        System.out.println(Arrays.toString(f.energyPerTurbine));
        System.out.println(Arrays.toString(f.turbineMoved));
        
        System.out.println();
        System.out.println();
        f.updateLocation(0, 0, 1);
        
        FarmLayout fCopy = (FarmLayout)deepCopy.DeepCopy.copy(f);
        
        
        System.out.println("\nOld evaluations:");
        
        w.valueOf(printEvaluationTimes, f, -1);
        System.out.print(f.totalEnergy+" ");
        System.out.println(Arrays.toString(f.energyPerTurbine));
        System.out.println(Arrays.toString(f.turbineMoved));
        f.updateLocation(1, 0, 350);
        w.valueOf(printEvaluationTimes, f, -1);
        System.out.print(f.totalEnergy+" ");
        System.out.println(Arrays.toString(f.energyPerTurbine));
        System.out.println(Arrays.toString(f.turbineMoved));
        f.updateLocation(2, 0, 700);
        w.valueOf(printEvaluationTimes, f, -1);
        System.out.print(f.totalEnergy+" ");
        System.out.println(Arrays.toString(f.energyPerTurbine));
        System.out.println(Arrays.toString(f.turbineMoved));
        
        
        System.out.println("\nImproved speed:");
        
        w.updateFast(printEvaluationTimes, fCopy);
        System.out.print(fCopy.totalEnergy+" ");
        System.out.println(Arrays.toString(fCopy.energyPerTurbine));
        System.out.println(Arrays.toString(fCopy.turbineMoved));
        fCopy.updateLocation(1, 0, 350);
        w.updateFast(printEvaluationTimes, fCopy);
        System.out.print(fCopy.totalEnergy+" ");
        System.out.println(Arrays.toString(fCopy.energyPerTurbine));
        System.out.println(Arrays.toString(fCopy.turbineMoved));
        fCopy.updateLocation(2, 0, 700);
        w.updateFast(printEvaluationTimes, fCopy);
        System.out.print(fCopy.totalEnergy+" ");
        System.out.println(Arrays.toString(fCopy.energyPerTurbine));
        System.out.println(Arrays.toString(fCopy.turbineMoved));
    }
    
    
    public void updateFast(boolean printEvaluationTimes, FarmLayout fCopy) {
        long startTime, stopTime;
        startTime = System.currentTimeMillis();
        
        boolean debugPrint = !true;

        int whichTurbineMoved = 0;
        for (int i = 0; i<fCopy.turbineMoved.length; i++) {
            if (fCopy.turbineMoved[i]==true) {
                whichTurbineMoved = i;
                break;
            }
        }

        for (int i = 0; i<fCopy.turbineMoved.length; i++) {
            if (debugPrint) System.out.println("### turbine "+i+" moved?"+fCopy.turbineMoved[i]);
            fCopy.energyPerTurbine[i] = energyForOneTurbine(fCopy.coordinates, i, false, fCopy, whichTurbineMoved);
            if (debugPrint) System.out.println("result:"+fCopy.energyPerTurbine[i]);
        }

        // reset all fields in this.turbineMoved
        for (int i = 0; i<fCopy.turbineMoved.length; i++) {
            fCopy.turbineMoved[i] = false;
        }
        
        double tempSum = 0;
        // add all the energies up
        for (double d:fCopy.energyPerTurbine) {
            tempSum += d;
        }
        
        // return sum
        fCopy.totalEnergy = tempSum;
        
        stopTime = System.currentTimeMillis();
        if (printEvaluationTimes) System.out.println("[time needed 'valueOf()':"+(stopTime-startTime)/1000.0+"s]");
    }

    
    
    static double[] config1 = {
        0,0,
        0,400,
        0,1000,
        500,0,
        500,500,
        500,1000,
        1000,0,
        1000,500,
        1000,1000
    };
    
    
    static double[] config2 = {
    7.07,8.52,
16.33,434.99,
0.01,817.34,
3.73,1244.82,
1.31,1656.06,
14.5,2094.7,
0.01,2508.46,
0,2932.58,
0.02,3342.88,
0.01,3783.56,
0,4181.22,
1.99,4602.01,
0.03,5000,
522.81,0.51,
523.75,410.31,
519.4,829.7,
526.2,1241.47,
528.88,1669.13,
553.91,2108.5,
527.5,2500.06,
526.32,2910.95,
538.51,3360.16,
516.06,3760.57,
545.37,4176.42,
512.51,4589.15,
520.9,5000,
1054.63,0.63,
1056.87,405.29,
1057.97,843.55,
1060.95,1262.62,
1067.4,1683.94,
1067.46,2107.8,
1047.05,2480.23,
1058.99,2917.23,
1070.53,3362.12,
1065.31,3741.79,
1050.65,4167.43,
1057.51,4612.51,
1059,5000,
1597.15,0.53,
1606.98,417.02,
1582.63,822.4,
1591.71,1236.67,
1569.24,1659.95,
1574.46,2061.86,
1600.64,2517.63,
1581.18,2891.87,
1586.64,3338.79,
1592.05,3759.23,
1587.32,4192.67,
1591.78,4589.35,
1566.49,4996.29,
2161.27,1.81,
2127.49,411.04,
2137.21,824.56,
2137.2,1252.56,
2144.82,1684.05,
2124.28,2076.13,
2147.58,2526.05,
2125.71,2931.34,
2117.57,3333.14,
2130.66,3779.88,
2109.88,4166.14,
2136.55,4611.54,
2111.65,4999.24,
2656.19,0,
2657.37,391.53,
2656.04,836.3,
2667.62,1259.87,
2647.04,1635.42,
2678.66,2116.16,
2654.41,2501.46,
2660.11,2930.16,
2662.43,3348.9,
2667.16,3757.39,
2670.67,4186.96,
2645.51,4586.69,
2647.69,4999.99,
3205.06,0.73,
3178.42,391.01,
3205.25,841.36,
3182.28,1234.92,
3209.77,1674.34,
3188.76,2064.45,
3215.48,2517.34,
3183.62,2904.02,
3216.12,3356.26,
3190.76,3754.05,
3198.6,4161.82,
3211.76,4603.16,
3186.54,4999.29,
3759.47,0.01,
3744.44,422.99,
3714.27,817.6,
3739.72,1251.24,
3721.11,1665.92,
3729.33,2084.28,
3740.03,2496.3,
3715.09,2894.37,
3740.68,3335.02,
3717.62,3734.35,
3745.83,4190.73,
3716.52,4578.82,
3726.6,5000,
4268.65,0.37,
4278.45,418.88,
4258.58,794.15,
4287.52,1277.95,
4268.06,1662.93,
4262.73,2077.51,
4276.15,2513.71,
4301.79,2949.84,
4269.78,3318.43,
4273.13,3754.35,
4278.9,4198.5,
4249.75,4579.94,
4251.41,4998.97,
4810.01,1.88,
4786.53,394.12,
4794.64,839.97,
4796.19,1235.84,
4802.21,1665.17,
4826.37,2095.28,
4794.49,2491.66,
4794.25,2908.82,
4804.06,3341.26,
4803.51,3748.25,
4810.91,4173.39,
4813.81,4605.19,
4795.35,4999.82,
5331.96,0,
5337.18,420.14,
5340.26,851.31,
5346.61,1265.45,
5330.32,1660.6,
5310.19,2049.38,
5336.67,2501.05,
5322.26,2923.77,
5346.53,3353.22,
5324.6,3742.26,
5327.2,4175.94,
5325.42,4580,
5327.73,4999.06,
5882.83,1.35,
5857.36,393.99,
5867.58,831.67,
5848.71,1219.32,
5860.26,1640.07,
5873.6,2074.74,
5872.05,2510.04,
5877.95,2937.72,
5879.53,3327.99,
5894.57,3780.1,
5861.26,4155.38,
5861.81,4593.64,
5871.72,4999.91,
6436.01,7.09,
6420.07,426.13,
6399.57,810.48,
6408.71,1262.89,
6388.22,1658.41,
6412.86,2108.18,
6387.08,2519.17,
6416.52,2924.49,
6395.91,3339.32,
6398.3,3741.36,
6398.37,4179.13,
6426.88,4605.61,
6389.89,4999.96,
6930.76,0.33,
6933.37,406.01,
6968.25,863.67,
6944.29,1259.21,
6956.63,1700.48,
6939.73,2093.85,
6928.55,2498.04,
6930.55,2875.77,
6930.73,3333.17,
6938.5,3740.95,
6920.69,4153.32,
6927.57,4593.34,
6911.28,4990.13,
7492.5,20.42,
7461.06,411.01,
7495.47,843.01,
7466.53,1238.79,
7479.91,1676.98,
7456.44,2068.57,
7499.5,2471.94,
7511.39,2894.5,
7488.32,3314.47,
7489.32,3748.34,
7496.72,4173.71,
7472,4579.44,
7468.86,5000,
7995.57,0,
7999.97,391.86,
7996.01,819.31,
8000,1253.54,
8000,1650.72
    };

    static double[] config3 = {
    1.23,0,
26.8,422.96,
3.93,793.88,
30.31,1288.42,
0,1662.17,
0.04,2104.57,
1.9,2543.16,
5.94,2919.15,
3.89,3323.42,
15.1,3754.48,
0.01,4203.53,
26.05,4607.77,
0.03,4980.58,
529.87,0.04,
555.8,440.19,
527.43,800.01,
557.02,1294.22,
524.49,1685.78,
520.43,2072.33,
490.64,2491.27,
530.26,2942.41,
544.86,3387.19,
522.47,3729.84,
549.61,4184.76,
528.9,4549.88,
533.34,4999.98,
1089.34,2.99,
1063.34,401.34,
1049.08,826.64,
1086.93,1251.8,
1051.94,1636.95,
1064.73,2098.56,
1049.48,2519.65,
1033.43,2907.85,
1084.38,3361.36,
1067.99,3755.82,
1077.19,4198.05,
1060.16,4592.49,
1101.25,4999.99,
1603.32,0.01,
1608.87,420.5,
1623.19,863.56,
1601.79,1231.23,
1628.27,1680.52,
1606.85,2095.27,
1577.98,2471.08,
1606.37,2884.67,
1588.56,3337.11,
1615.59,3786.76,
1596.26,4173.94,
1603.15,4549.9,
1572.75,4990.64,
2123.4,0.03,
2162.23,401.87,
2122.54,816.03,
2151.65,1205.46,
2159.9,1670.75,
2139.97,2079.96,
2153.03,2512.32,
2128.33,2904.12,
2154.76,3374.36,
2115.02,3785.98,
2125.07,4193.74,
2140.19,4619.65,
2120.17,4997.85,
2672.44,4.04,
2651.29,398.53,
2657.31,835.93,
2634.69,1218.48,
2644.01,1654.5,
2675.27,2085.86,
2678.96,2492.71,
2661.65,2895.96,
2677.55,3327.12,
2654.06,3734.4,
2683.2,4166.34,
2653.91,4570.47,
2682.66,4973.36,
3208.71,0,
3189.94,414.62,
3215.83,865.71,
3185.03,1230.78,
3204.01,1712.74,
3184.59,2097.74,
3210.26,2491.04,
3187.17,2894.34,
3216,3353.42,
3188.95,3700.05,
3176.92,4199.75,
3211.08,4599.3,
3174.62,4997.17,
3744.91,0.01,
3741.22,413.59,
3718.61,817.51,
3743.84,1269.44,
3715.98,1667.02,
3743.95,2078.95,
3716.75,2519.2,
3739.46,2941.85,
3745.51,3316.76,
3723.76,3759.7,
3730.89,4180.54,
3759.15,4596.62,
3728.61,5000,
4279.94,7.11,
4261.12,387.36,
4266.87,828.43,
4267.71,1249.28,
4293.14,1697.74,
4271.35,2094.06,
4247.21,2477.77,
4280.46,2969.57,
4252.1,3328.5,
4278.15,3741.07,
4263.95,4200.2,
4237.27,4579.85,
4265.46,4980.79,
4802.31,0.85,
4805.81,419.64,
4782.06,803.29,
4810.7,1264.49,
4781.36,1635.02,
4807.06,2074.73,
4786.12,2485.07,
4812.77,2941.83,
4790.46,3303.8,
4820.85,3793.42,
4799.37,4153.22,
4776.09,4560.46,
4789.94,4999.55,
5337.66,0.07,
5357.68,428.99,
5342.94,842.58,
5346.02,1236.34,
5360.04,1676.81,
5345.1,2116.38,
5352.87,2478.9,
5335.39,2915.3,
5362.07,3360.81,
5331.02,3724.36,
5357.32,4200.3,
5331.56,4579.74,
5343.16,5000,
5855.15,0.1,
5887.62,430.88,
5853.85,808.86,
5881.99,1268.61,
5861.07,1674.53,
5879.98,2096.8,
5849.95,2455.92,
5862.16,2931.47,
5895.68,3369.63,
5864.63,3739.1,
5882.14,4197.03,
5855.47,4538.11,
5867.19,4999.68,
6412,0,
6382.77,425.68,
6418.37,829.76,
6415.03,1223.11,
6393.22,1675.21,
6421.42,2097.97,
6387.77,2513.75,
6410.99,2910.45,
6391.85,3329.97,
6406.45,3778.8,
6379.46,4128.9,
6408.63,4603.62,
6417.28,4999.99,
6954.04,1,
6916.17,401.99,
6952.95,809.46,
6925.49,1251.66,
6964.94,1695.31,
6969.5,2069.66,
6941.66,2521.44,
6968.3,2905.88,
6946.62,3329.47,
6936.84,3738.95,
6919.87,4164.58,
6947.04,4619.54,
6918.38,4995.3,
7503.49,20.66,
7484.81,397.58,
7494.02,826.52,
7480.55,1258.06,
7497.31,1675.77,
7480.49,2082.27,
7501.98,2501.87,
7480.05,2865.52,
7463.09,3304,
7506,3710.88,
7471.98,4138.42,
7482.37,4595.15,
7466.62,4987.73,
7999.87,0,
8000,402.03,
7993.37,796.51,
7999.99,1225.27,
7975.31,1709.77
    };

    static double[] config4 = {
    0,0,
0,1111.11,
0,2222.22,
0,3333.33,
0,4444.44,
0,5555.56,
0,6666.67,
0,7777.78,
0,8888.89,
0,10000,
0,11111.11,
0,12222.22,
0,13333.33,
0,14444.44,
0,15555.56,
0,16666.67,
0,17777.78,
0,18888.89,
0,20000,
1000,0,
1000,1111.11,
1000,2222.22,
1000,3333.33,
1000,4444.44,
1000,5555.56,
1000,6666.67,
1000,7777.78,
1000,8888.89,
1000,10000,
1000,11111.11,
1000,12222.22,
1000,13333.33,
1000,14444.44,
1000,15555.56,
1000,16666.67,
1000,17777.78,
1000,18888.89,
1000,20000,
2000,0,
2000,1111.11,
2000,2222.22,
2000,3333.33,
2000,4444.44,
2000,5555.56,
2000,6666.67,
2000,7777.78,
2000,8888.89,
2000,10000,
2000,11111.11,
2000,12222.22,
2000,13333.33,
2000,14444.44,
2000,15555.56,
2000,16666.67,
2000,17777.78,
2000,18888.89,
2000,20000,
3000,0,
3000,1111.11,
3000,2222.22,
3000,3333.33,
3000,4444.44,
3000,5555.56,
3000,6666.67,
3000,7777.78,
3000,8888.89,
3000,10000,
3000,11111.11,
3000,12222.22,
3000,13333.33,
3000,14444.44,
3000,15555.56,
3000,16666.67,
3000,17777.78,
3000,18888.89,
3000,20000,
4000,0,
4000,1111.11,
4000,2222.22,
4000,3333.33,
4000,4444.44,
4000,5555.56,
4000,6666.67,
4000,7777.78,
4000,8888.89,
4000,10000,
4000,11111.11,
4000,12222.22,
4000,13333.33,
4000,14444.44,
4000,15555.56,
4000,16666.67,
4000,17777.78,
4000,18888.89,
4000,20000,
5000,0,
5000,1111.11,
5000,2222.22,
5000,3333.33,
5000,4444.44,
5000,5555.56,
5000,6666.67,
5000,7777.78,
5000,8888.89,
5000,10000,
5000,11111.11,
5000,12222.22,
5000,13333.33,
5000,14444.44,
5000,15555.56,
5000,16666.67,
5000,17777.78,
5000,18888.89,
5000,20000,
6000,0,
6000,1111.11,
6000,2222.22,
6000,3333.33,
6000,4444.44,
6000,5555.56,
6000,6666.67,
6000,7777.78,
6000,8888.89,
6000,10000,
6000,11111.11,
6000,12222.22,
6000,13333.33,
6000,14444.44,
6000,15555.56,
6000,16666.67,
6000,17777.78,
6000,18888.89,
6000,20000,
7000,0,
7000,1111.11,
7000,2222.22,
7000,3333.33,
7000,4444.44,
7000,5555.56,
7000,6666.67,
7000,7777.78,
7000,8888.89,
7000,10000,
7000,11111.11,
7000,12222.22,
7000,13333.33,
7000,14444.44,
7000,15555.56,
7000,16666.67,
7000,17777.78,
7000,18888.89,
7000,20000,
8000,0,
8000,1111.11,
8000,2222.22,
8000,3333.33,
8000,4444.44,
8000,5555.56,
8000,6666.67,
8000,7777.78,
8000,8888.89,
8000,10000,
8000,11111.11,
8000,12222.22,
8000,13333.33,
8000,14444.44,
8000,15555.56,
8000,16666.67,
8000,17777.78,
8000,18888.89,
8000,20000,
9000,0,
9000,1111.11,
9000,2222.22,
9000,3333.33,
9000,4444.44,
9000,5555.56,
9000,6666.67,
9000,7777.78,
9000,8888.89,
9000,10000,
9000,11111.11,
9000,12222.22,
9000,13333.33,
9000,14444.44,
9000,15555.56,
9000,16666.67,
9000,17777.78,
9000,18888.89,
9000,20000,
10000,0,
10000,1111.11,
10000,2222.22,
10000,3333.33,
10000,4444.44,
10000,5555.56,
10000,6666.67,
10000,7777.78,
10000,8888.89,
10000,10000
    };
}